import os
import uuid

import numpy as np
import pytest

from libertem.common.shape import Shape
from libertem.common.math import prod
from libertem.io.dataset.base.exceptions import DataSetException
from libertem.io.dataset.base.tiling_scheme import TilingScheme
from libertem.udf.base import UDF
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.dm_single import SingleDMDataSet
from libertem.contrib.convert_transposed import convert_dm4_transposed

try:
    import hyperspy.api as hs
except ModuleNotFoundError:
    hs = None

from utils import ValidationUDF, _mk_random, dataset_correction_verification, get_testdata_path


DM_TESTDATA_PATH = os.path.join(get_testdata_path(), 'dm')
HAVE_DM_TESTDATA = os.path.exists(DM_TESTDATA_PATH)


@pytest.fixture(scope='module')
def dm4_c_order_4d_path():
    return os.path.join(DM_TESTDATA_PATH, '4d_data_c_order_zen4307783.dm4')


@pytest.fixture(scope='module')
def dm4_f_order_3d_path():
    return os.path.join(DM_TESTDATA_PATH, '3d_data_f_order_zen2580185.dm4')


@pytest.fixture(scope='module')
def dm3_3dstack_path():
    return os.path.join(DM_TESTDATA_PATH, '3D', 'alpha-50_obj.dm3')


@pytest.fixture(scope='module')
def dm4_2dimage_path():
    return os.path.join(DM_TESTDATA_PATH, '2018-7-17 15_29_0000.dm4')


@pytest.fixture(scope='module')
def dm4_c_order_4d_raw(dm4_c_order_4d_path):
    return hs.load(dm4_c_order_4d_path).data


@pytest.fixture(scope='module')
def dm4_f_order_3d_raw(dm4_f_order_3d_path):
    return hs.load(dm4_f_order_3d_path).data


@pytest.fixture(scope='module')
def dm3_3dstack_raw(dm3_3dstack_path):
    return hs.load(dm3_3dstack_path).data


@pytest.fixture(scope='module')
def dm4_2dimage_raw(dm4_2dimage_path):
    return hs.load(dm4_2dimage_path).data


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
def test_3d_f_order(lt_ctx_fast, dm4_f_order_3d_path):
    with pytest.raises(DataSetException):
        lt_ctx_fast.load('dm', dm4_f_order_3d_path)


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
@pytest.mark.slow
def test_3d_f_order_force(lt_ctx_fast, dm4_f_order_3d_path, dm4_f_order_3d_raw):
    ds = lt_ctx_fast.load('dm', dm4_f_order_3d_path, force_c_order=True)
    assert ds.shape == Shape((3710, 146, 228), sig_dims=1)
    res = lt_ctx_fast.run_udf(ds, SumSigUDF())
    # best we can do!
    assert np.allclose(res['intensity'].data.sum(), dm4_f_order_3d_raw.sum())


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
@pytest.mark.slow
def test_3d_c_order(lt_ctx_fast, dm3_3dstack_path, dm3_3dstack_raw):
    ds = lt_ctx_fast.load('dm', dm3_3dstack_path)
    assert ds.shape == Shape((20, 3838, 3710), sig_dims=2)
    res = lt_ctx_fast.run_udf(ds, SumSigUDF())
    assert np.allclose(res['intensity'].data, dm3_3dstack_raw.sum(axis=(-2, -1)))


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
@pytest.mark.slow
def test_2d_c_order(lt_ctx_fast, dm4_2dimage_path, dm4_2dimage_raw):
    ds = lt_ctx_fast.load('dm', dm4_2dimage_path)
    assert ds.shape == Shape((1, 3838, 3710), sig_dims=2)
    res = lt_ctx_fast.run_udf(ds, SumSigUDF())
    assert np.allclose(res['intensity'].data.squeeze(), dm4_2dimage_raw.sum(axis=(-2, -1)))


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
@pytest.mark.slow
def test_4d_c_order(lt_ctx_fast, dm4_c_order_4d_path, dm4_c_order_4d_raw):
    ds = lt_ctx_fast.load('dm', dm4_c_order_4d_path)
    assert isinstance(ds, SingleDMDataSet)
    assert ds.shape.to_tuple() == (213, 342, 128, 128)
    assert ds.dtype == np.uint8
    res = lt_ctx_fast.run_udf(ds, SumSigUDF())
    assert np.allclose(res['intensity'].data, dm4_c_order_4d_raw.sum(axis=(-2, -1)))


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
@pytest.mark.slow
def test_comparison_raw(lt_ctx_fast, dm4_c_order_4d_path, dm4_c_order_4d_raw):
    ds = lt_ctx_fast.load('dm', dm4_c_order_4d_path)
    assert ds.meta.shape.to_tuple() == dm4_c_order_4d_raw.shape
    flat_data = dm4_c_order_4d_raw.reshape(-1, *dm4_c_order_4d_raw.shape[-2:])
    udf = ValidationUDF(reference=flat_data)
    lt_ctx_fast.run_udf(udf=udf, dataset=ds)


# Remaining tests are on mockfile objects rather than the large, real files


class MockFileDM:
    def __init__(self, shapes, offsets, dtypes, orders):
        self.dataShape = []
        self.xSize = []
        self.ySize = []
        self.zSize = []
        self.zSize2 = []
        self.dataOffset = []
        self.dataType = []
        self.allTags = {}
        self.thumbnail = False
        self.numObjects = len(shapes)

        c_shape_order = (self.zSize2,
                         self.zSize,
                         self.ySize,
                         self.xSize)
        f_shape_order = (self.ySize,
                         self.xSize,
                         self.zSize2,
                         self.zSize)

        for ds_idx, (shape, offset, dtype, order) in enumerate(zip(shapes,
                                                                   offsets,
                                                                   dtypes,
                                                                   orders),
                                                               start=1):
            self.dataShape.append(len(shape))
            self.dataOffset.append(offset)
            self.dataType.append(dtype)
            if order == 'C':
                key = f'ImageList.{ds_idx}.ImageTags.Meta Data.Data Order Swapped'
                self.allTags[key] = 1
                shape_unpack = zip(shape, c_shape_order)
            else:
                shape_unpack = zip(shape, f_shape_order)
            key = f'ImageList.{ds_idx}.ImageTags.Meta Data.Format'
            self.allTags[key] = 'Diffraction image'
            for s, dim in shape_unpack:
                dim.append(s)

    def __call__(self, path, on_memory=True):
        return self

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        ...

    def _DM2NPDataType(self, dtype):
        if dtype is None:
            raise OSError('no dtype')
        return dtype


def generate_dm4_mockfile(filename, shape, dtype,
                          c_order=False, header=1000):
    array = _mk_random(shape.to_tuple(), dtype=dtype)
    sig_dims = shape.sig.dims

    order = 'C' if c_order else 'F'

    with open(filename, 'ab') as fp:
        fp.write(b'0' * header)
        offset = fp.tell()
        if c_order:
            array.tofile(fp)
        else:
            dims = tuple(range(shape.dims))
            f_axis_order = dims[-sig_dims:] + dims[:-sig_dims]
            np.transpose(array, f_axis_order).tofile(fp)

    return array, order, offset


def _make_mock_array(datadir, shape, dtype, c_order):
    suffix = str(uuid.uuid1())[:4]
    filename = os.path.join(datadir, f'dm4_mock_f_{suffix}.dm4')
    array, order, offset = generate_dm4_mockfile(filename,
                                                 shape,
                                                 dtype,
                                                 c_order=c_order)
    mock_fileDM = MockFileDM((shape,), (offset,), (dtype,), (order,))
    return (array, filename), mock_fileDM


@pytest.fixture(scope='module')
def dm4_mockfile_f(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    shape = Shape((4, 8, 16, 24), 2)
    dtype = np.float32
    return _make_mock_array(datadir, shape, dtype, False)


@pytest.fixture(scope='module')
def dm4_mockfile_c(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    shape = Shape((4, 8, 16, 24), 2)
    dtype = np.float32
    return _make_mock_array(datadir, shape, dtype, True)


def _patch_filedm(monkeypatch, dm_obj):
    monkeypatch.setattr('ncempy.io.dm.fileDM', dm_obj)


def test_f_order_raises(monkeypatch, dm4_mockfile_f, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_f
    _patch_filedm(monkeypatch, mock_fileDM)

    with pytest.raises(DataSetException):
        lt_ctx_fast.load('dm', filename)


def test_comparison(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)
    assert ds.meta.shape.to_tuple() == array.shape
    flat_data = array.reshape(-1, *array.shape[-2:])
    udf = ValidationUDF(reference=flat_data)
    lt_ctx_fast.run_udf(udf=udf, dataset=ds)


def test_comparison_roi(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)
    assert ds.meta.shape.to_tuple() == array.shape

    roi = np.random.choice(
        [True, False],
        size=tuple(ds.shape.nav),
        p=[0.5, 0.5]
    )

    flat_data = array.reshape(-1, *array.shape[-2:])
    udf = ValidationUDF(reference=flat_data[roi.ravel()])
    lt_ctx_fast.run_udf(udf=udf, dataset=ds, roi=roi)


def test_many_tiles(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)
    ds.set_num_cores(8)
    _, _, sy, sx = array.shape
    flat_data = array.reshape(-1, sy, sx)
    # depth 5, height 3, divides neither flat_nav or sy evenly
    tileshape = Shape((5, 3, sx), sig_dims=2)
    tiling_scheme = TilingScheme.make_for_shape(tileshape, ds.meta.shape, intent='tile')
    parts = [*ds.get_partitions()]
    assert len(parts) > 1
    ds_visited = np.zeros(flat_data.shape, dtype=int)
    for part in parts:
        for tile in part.get_tiles(tiling_scheme):
            tile_slice = tile.tile_slice
            assert np.allclose(tile_slice.get(flat_data), tile.data)
            ds_visited[tile_slice.get()] += 1
    assert (ds_visited == 1).all()


class SumFrameUDF(UDF):
    def get_result_buffers(self):
        return {'sum': self.buffer('nav')}

    def process_frame(self, frame):
        self.results.sum[:] += frame.sum()


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_process_frame(monkeypatch, dm4_mockfile_c, lt_ctx_fast, with_roi):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)
    result = array.sum(axis=(2, 3))

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
        result[np.logical_not(roi)] = np.nan
    else:
        roi = None

    res = lt_ctx_fast.run_udf(dataset=ds, udf=SumFrameUDF(), roi=roi)
    if roi is not None:
        assert (np.isnan(res['sum'].data) == np.logical_not(roi)).all()
    assert np.allclose(res['sum'].data, result, equal_nan=True)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_corrections_default(monkeypatch, dm4_mockfile_c, lt_ctx_fast, with_roi):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx_fast)


def test_macrotile_normal(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)
    ds.set_num_cores(4)

    ps = ds.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    macrotile = p2.get_macrotile()
    assert macrotile.tile_slice.shape == p2.shape
    assert macrotile.tile_slice.origin[0] == p2._start_frame


def test_macrotile_roi(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    ds = lt_ctx_fast.load('dm', filename)

    roi = np.zeros(ds.shape.nav, dtype=bool)
    roi[0, 5] = 1
    roi[1, 1] = 1
    p = next(ds.get_partitions())
    macrotile = p.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == (2, 16, 24)


def test_positive_sync_offset(monkeypatch, dm4_mockfile_c, lt_ctx):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    udf = SumSigUDF()
    sync_offset = 2

    ds_no_offset = lt_ctx.load('dm', filename)
    ds_with_offset = SingleDMDataSet(
        path=filename,
        sync_offset=sync_offset,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    # t0 = next(p0.get_tiles(tiling_scheme))
    # Not guaranteed as FortranReader can emit tiles out-of-order
    # assert tuple(t0.tile_slice.origin) == (0, 0, 0)

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (24, 0, 0)
    assert p.slice.shape[0] == 8

    result = lt_ctx.run_udf(dataset=ds_no_offset, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


def test_negative_sync_offset(monkeypatch, dm4_mockfile_c, lt_ctx):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    udf = SumSigUDF()
    sync_offset = -2

    ds_no_offset = lt_ctx.load('dm', filename)
    ds_with_offset = SingleDMDataSet(
        path=filename,
        sync_offset=sync_offset,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    # t0 = next(p0.get_tiles(tiling_scheme))
    # Not guaranteed as FortranReader can emit tiles out-of-order
    # assert tuple(t0.tile_slice.origin) == (0, 0, 0)

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (24, 0, 0)
    assert p.slice.shape[0] == 8

    result = lt_ctx.run_udf(dataset=ds_no_offset, udf=udf)
    result = result['intensity'].raw_data[:ds_no_offset._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


def test_nav_reshaping_c_error(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    nav_shape = tuple(d + 1 for d in array.shape[:2])
    with pytest.raises(DataSetException):
        lt_ctx_fast.load('dm', filename, nav_shape=nav_shape)


def test_nav_reshaping_c(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    # read only part of the nav space
    manual_nav_shape = tuple(d - 1 for d in array.shape[:2])
    ds = lt_ctx_fast.load('dm', filename, nav_shape=manual_nav_shape)
    res = lt_ctx_fast.run_udf(dataset=ds, udf=SumSigUDF())
    result_array = res['intensity'].data

    _, _, sy, sx = array.shape
    nframes = prod(manual_nav_shape)
    partial_sum = array.reshape(-1, sy, sx).sum(axis=(1, 2))[:nframes]
    assert np.allclose(result_array, partial_sum.reshape(manual_nav_shape))


def test_sig_reshaping_c_error(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    sig_shape = tuple(d + 1 for d in array.shape[2:])
    with pytest.raises(DataSetException):
        lt_ctx_fast.load('dm', filename, sig_shape=sig_shape)


def test_sig_reshaping_c(monkeypatch, dm4_mockfile_c, lt_ctx_fast):
    (array, filename), mock_fileDM = dm4_mockfile_c
    _patch_filedm(monkeypatch, mock_fileDM)

    # read only part of each sig space
    manual_sig_shape = tuple(d - 1 for d in array.shape[:2])
    ds = lt_ctx_fast.load('dm', filename, sig_shape=manual_sig_shape)
    res = lt_ctx_fast.run_udf(dataset=ds, udf=SumSigUDF())
    result_array = res['intensity'].data

    ny, nx = array.shape[:2]
    flat_array = array.ravel()
    nel = ny * nx * prod(manual_sig_shape)
    partial_array = flat_array[:nel].reshape(array.shape[:2] + manual_sig_shape)
    partial_sum = partial_array.sum(axis=(2, 3))
    assert np.allclose(result_array, partial_sum)


@pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="No DM4 test data")
def test_convert_not_transposed_raises(lt_ctx, dm4_c_order_4d_path):
    with pytest.raises(DataSetException):
        convert_dm4_transposed(
            dm4_c_order_4d_path,
            'out.npy',
            ctx=lt_ctx,
        )


def test_convert_f_ordered(monkeypatch, dm4_mockfile_f, lt_ctx, tmpdir_factory):
    tdir = tmpdir_factory.mktemp('convert_transposed')

    # the mockfile array in memory is already c-ordered
    (array, filename), mock_fileDM = dm4_mockfile_f
    _patch_filedm(monkeypatch, mock_fileDM)

    converted_path = os.path.join(tdir, 'out_dm4_f.npy')
    convert_dm4_transposed(
        filename,
        converted_path,
        ctx=lt_ctx,
    )

    ds_npy = lt_ctx.load('npy', path=converted_path)
    pick_a = lt_ctx.create_pick_analysis(ds_npy, 4, 3)
    pick_frame = lt_ctx.run(pick_a).intensity.raw_data
    # flipped coords because pick_a takes x, y arguments
    assert np.allclose(pick_frame, array[3, 4, :, :])



import tornado
import logging

from .state import SharedState

log = logging.getLogger(__name__)


class ShutdownHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def delete(self):
        log.info("Handling shutdown button")
        self.state.executor_state.shutdown()
        await self.finish({
            "status": "ok",
            "messageType": "SERVER_SHUTDOWN"
        })
        tornado.ioloop.IOLoop.current().stop()



# Imported for backwards compatibility, refs #1031
from libertem.common.async_utils import (  # NOQA: F401
    MyStopIteration, sync_to_async, async_generator, run_agen_get_last,
    run_gen_get_last, AsyncGenToQueueThread, async_to_sync_generator,
    SyncGenToQueueThread, async_generator_eager, adjust_event_loop_policy
)



import logging
from typing import Optional
import contextlib

import numpy as np
from sparseconverter import ArrayBackend

from .fileset import FileSet

log = logging.getLogger(__name__)


class IOBackend:
    registry: dict[str, type["IOBackend"]] = {}
    id_: Optional[str] = None

    def __init_subclass__(cls, id_: Optional[str] = None):
        super().__init_subclass__()
        if id_ is not None:
            cls.registry[id_] = cls
            cls.id_ = id_

    @classmethod
    def get_cls_by_id(cls, id_):
        return cls.registry.get(id_)

    @classmethod
    def get_supported(cls):
        return [
            k
            for k in cls.registry.keys()
            if cls.registry[k].platform_supported()
        ]

    def __init__(self):
        pass

    def get_impl(self) -> 'IOBackendImpl':
        raise NotImplementedError()

    @classmethod
    def platform_supported(cls):
        return True

    @classmethod
    def from_json(cls, msg):
        """
        Construct an instance from the already-decoded `msg`.
        """
        raise NotImplementedError()


class IOBackendImpl:
    def __init__(self):
        pass

    @contextlib.contextmanager
    def open_files(self, fileset: FileSet):
        """
        Open files, yielding a list of implementation-specific file objects
        representing these open files.

        Parameters
        ----------
        fileset : FileSet
            [description]
        """
        raise NotImplementedError()

    def need_copy(
        self, decoder, roi, native_dtype, read_dtype, tiling_scheme=None, fileset=None,
        sync_offset=0, corrections=None,
    ) -> bool:
        # checking conditions in which "straight mmap" is not possible
        # straight mmap means our dataset can just return views into the underlying mmap object
        # as tiles and use them as they are in the UDFs

        # 1) if a roi is given, straight mmap doesn't work because there are gaps in the navigation
        # axis:
        if roi is not None:
            log.debug("have roi, need copy")
            return True

        # 2) if we need to decode data, or do dtype conversion, we can't return
        # views into the underlying file:
        if self._need_decode(decoder, native_dtype, read_dtype):
            log.debug("have decode, need copy")
            return True

        # 3) if we have less number of frames per file than tile depth, we need to copy
        if tiling_scheme and fileset:
            fileset_arr = fileset.get_as_arr()
            if np.min(fileset_arr[:, 1] - fileset_arr[:, 0]) < tiling_scheme.depth:
                log.debug("too large for fileset, need copy")
                return True

        # 4) if we apply corrections, we need to copy
        if corrections is not None and corrections.have_corrections():
            log.debug("have corrections, need copy")
            return True

        # 5) if a negative offset is given, we need to copy
        if sync_offset < 0:
            log.debug("negative offset is set, need copy")
            return True

        return False

    def get_max_io_size(self):
        return 2**20  # default: 1MiB blocks

    def _need_decode(self, decoder, native_dtype, read_dtype):
        # FIXME: even with dtype "mismatch", we can possibly do dtype
        # conversion, if the tile size is small enough! maybe benchmark this
        # vs. _get_tiles_w_copy?
        if native_dtype != read_dtype:
            return True
        if decoder is not None:
            return True
        return False

    def preprocess(self, data, tile_slice, corrections):
        if corrections is None:
            return
        corrections.apply(data, tile_slice)

    def get_tiles(
        self, tiling_scheme, fileset, read_ranges, roi, native_dtype, read_dtype, decoder,
        sync_offset, corrections, array_backend: ArrayBackend,
    ):
        """
        Read tiles from `fileset`, as specified by the parameters.

        Usually, this is used to read the data for a single partition.

        Parameters
        ----------

        tiling_scheme : TilingScheme
            Specifies how the tiles should be shaped

        fileset : FileSet
            The files that should be read from. Note that the order in the `FileSet` is important,
            it must match the indices on the `read_ranges`.

        read_ranges : np.ndarray
            Read ranges, as generated by :meth:`FileSet.get_read_ranges`

        roi : np.ndarray
            Boolean array specifying which data should be read

        array_backend : ArrayBackend
            Array backend to use for reading the data

            .. versionadded:: 0.11.0

        native_dtype : np.dtype
            The native on-disk data type. If there is no direct match to
            a numpy dtype, specify the closest dtype.

        read_dtype : np.dtype
            The data dtype into which the data is converted when reading

        decoder : libertem.io.dataset.base.Decoder

        sync_offset : int
            if positive, number of frames to skip from the start
            if negative, number of blank frames to insert at the start

        corrections
            A set of corrections to apply in a preprocesing step
        """
        raise NotImplementedError()



from functools import partial

import numba
import pytest
import dask
import dask.array as da
import numpy as np
from libertem.executor.delayed import DelayedUDFRunner

from libertem.udf.base import UDF
from libertem.udf.masks import ApplyMasksUDF

from libertem.io.dataset.memory import MemoryDataSet
from libertem.executor.utils.dask_inplace import DaskInplaceWrapper

from utils import _mk_random


def test_inplace_wrapper():
    data = _mk_random((55, 55), dtype=np.float32)
    dask_wrapped = DaskInplaceWrapper(da.from_array(data.copy()))
    dask_wrapped.set_slice(np.s_[2, :])
    # this used to raise `NotImplementedError` but should now work everywhere:
    dask_wrapped[np.s_[3]] = 55.


class CountUDF(UDF):
    def process_tile(self, tile):
        self.results.count[:] = np.ones(tile.shape[:1], dtype=np.int32)

    def get_result_buffers(self):
        return {
            'count': self.buffer(kind='nav', dtype=np.int32),
            'total': self.buffer(
                kind='single',
                dtype=np.int32,
                use='result_only'
            ),
        }

    def get_results(self):
        return {
            'total': self.results.count.sum()
        }


class StrUDF(UDF):
    def process_tile(self, tile):
        self.results.str[:] = np.full(tile.shape[:1], 'a', dtype=np.dtype('S1'))

    def get_result_buffers(self):
        return {'str': self.buffer(kind='nav', dtype=np.dtype('S1'))}


class MySumUDF(UDF):
    def get_result_buffers(self):
        return {
            'intensity': self.buffer(kind='sig', dtype=self.meta.input_dtype),
        }

    def process_tile(self, tile):
        self.results.intensity[:] += np.sum(tile, axis=0)

    def merge(self, dest, src):
        dest.intensity[:] += src.intensity


class MySumMergeUDF(MySumUDF):
    def merge_all(self, ordered_results):
        intensity = np.stack([b.intensity for b in ordered_results.values()]).sum(axis=0)
        return {
            'intensity': intensity
        }


class OnlyDaskSumUDF(UDF):
    def get_result_buffers(self):
        return {
            'intensity': self.buffer(kind='sig', dtype=self.meta.input_dtype)
        }

    def process_tile(self, tile):
        self.results.intensity[:] += np.sum(tile, axis=0)

    def merge_all(self, ordered_results):
        intensity = np.stack([b.intensity for b in ordered_results.values()]).sum(axis=0)
        return {
            'intensity': intensity
        }


@numba.njit(cache=True)
def _process_tile(dest, tile):
    dest += np.sum(tile, axis=0)


@numba.njit(cache=True)
def _merge(dest, src):
    dest += src


class NumbaSumUDF(UDF):
    def get_result_buffers(self):
        return {
            'intensity': self.buffer(kind='sig', dtype=self.meta.input_dtype)
        }

    def process_tile(self, tile):
        _process_tile(self.results.intensity[:], tile)

    def merge(self, dest, src):
        _merge(dest.intensity[:], src.intensity)


class NumbaSumMergeUDF(MySumUDF):
    def merge_all(self, ordered_results):
        intensity = np.stack([b.intensity for b in ordered_results.values()]).sum(axis=0)
        return {
            'intensity': intensity
        }


class MySumSigUDF(UDF):
    def get_result_buffers(self):
        return {
            'intensity': self.buffer(
                kind="nav", dtype=self.meta.input_dtype
            ),
        }

    # Make sure we don't have default merge so that there is no default merge_all
    def merge(self, dest, src):
        return super().merge(dest, src)

    def process_tile(self, tile):
        self.results.intensity[:] += np.sum(tile, axis=tuple(range(1, len(tile.shape))))


class MySumSigMergeUDF(MySumSigUDF):
    def merge_all(self, ordered_results):
        intensity = np.concatenate([b.intensity for b in ordered_results.values()])
        return {
            'intensity': intensity
        }


def get_dataset(ctx, shape, tileshape, num_partitions, sig_dims, use_roi=False):
    data = _mk_random(shape)
    dataset = MemoryDataSet(data=data, tileshape=tileshape,
                            num_partitions=num_partitions, sig_dims=sig_dims)
    dataset.initialize(ctx)
    if use_roi:
        roi = np.random.choice([True, False], dataset._shape.nav)
    else:
        roi = None
    return {'data': data, 'dataset': dataset, 'roi': roi}


def ds_dims(dataset):
    ds_shape = dataset._shape
    n_sig_dims = ds_shape.sig.dims
    nav_dims = tuple(range(ds_shape.dims - n_sig_dims))
    sig_dims = tuple(range(ds_shape.nav.dims, ds_shape.dims))
    return ds_shape, nav_dims, sig_dims


def count(udf_params, ds_dict):
    data = ds_dict['data']
    dataset = ds_dict['dataset']
    roi = ds_dict.get('roi', None)
    ds_shape, nav_dims, sig_dims = ds_dims(dataset)

    flat_nav_data, flat_sig_dims = flatten_with_roi(data, roi, ds_shape)

    udf = CountUDF(**udf_params)
    naive_result = np.ones(flat_nav_data.shape[:1], dtype=np.int32)
    naive_result = fill_nav_with_roi(ds_shape, naive_result, roi, fill=0)
    return {
        'udf': udf,
        'naive_result': {
            'count': naive_result,
            'total': naive_result.sum(),
        }
    }


def string(udf_params, ds_dict):
    data = ds_dict['data']
    dataset = ds_dict['dataset']
    roi = ds_dict.get('roi', None)
    ds_shape, nav_dims, sig_dims = ds_dims(dataset)

    flat_nav_data, flat_sig_dims = flatten_with_roi(data, roi, ds_shape)

    udf = StrUDF(**udf_params)
    naive_result = np.full(flat_nav_data.shape[:1], 'a', dtype=np.dtype('S1'))
    naive_result = fill_nav_with_roi(ds_shape, naive_result, roi, fill='')
    return {'udf': udf, 'naive_result': {'str': naive_result}}


def _navsum(udf_class, udf_params, ds_dict):
    data = ds_dict['data']
    dataset = ds_dict['dataset']
    roi = ds_dict.get('roi', None)
    ds_shape, nav_dims, sig_dims = ds_dims(dataset)

    flat_nav_data, _ = flatten_with_roi(data, roi, ds_shape)

    udf = udf_class(**udf_params)
    naive_result = flat_nav_data.sum(axis=0)
    return {'udf': udf, 'naive_result': {'intensity': naive_result}}


def navsum(udf_params, ds_dict):
    return _navsum(MySumUDF, udf_params, ds_dict)


def navsumdask(udf_params, ds_dict):
    return _navsum(MySumMergeUDF, udf_params, ds_dict)


def numba_navsum(udf_params, ds_dict):
    return _navsum(NumbaSumUDF, udf_params, ds_dict)


def numba_navsumdask(udf_params, ds_dict):
    return _navsum(NumbaSumMergeUDF, udf_params, ds_dict)


def _sigsum(udf_class, udf_params, ds_dict):
    data = ds_dict['data']
    dataset = ds_dict['dataset']
    roi = ds_dict.get('roi', None)
    ds_shape, nav_dims, sig_dims = ds_dims(dataset)

    flat_nav_data, flat_sig_dims = flatten_with_roi(data, roi, ds_shape)

    udf = udf_class(**udf_params)
    naive_result = flat_nav_data.sum(axis=flat_sig_dims)
    naive_result = fill_nav_with_roi(ds_shape, naive_result, roi)

    return {'udf': udf, 'naive_result': {'intensity': naive_result}}


def sigsum(udf_params, ds_dict):
    return _sigsum(MySumSigUDF, udf_params, ds_dict)


def sigsumdask(udf_params, ds_dict):
    return _sigsum(MySumSigMergeUDF, udf_params, ds_dict)


def _mask_fac(mask_slice, mask_value, ds_shape):
    m = np.zeros(ds_shape.sig)
    m[mask_slice] = mask_value
    return m


def mask(udf_params, ds_dict):
    data = ds_dict['data']
    dataset = ds_dict['dataset']
    roi = ds_dict.get('roi', None)
    ds_shape, nav_dims, sig_dims = ds_dims(dataset)

    flat_nav_data, flat_sig_dims = flatten_with_roi(data, roi, ds_shape)

    mask_slice = udf_params.pop('slices', [np.s_[-1, -1]])
    mask_value = udf_params.pop('values', [1.3])

    factories = []
    for sl, val in zip(mask_slice, mask_value):
        factories.append(partial(_mask_fac, sl, val, ds_shape))

    udf = ApplyMasksUDF(mask_factories=factories)

    results = []
    for fac in factories:
        _mask = fac()
        masked_result = (flat_nav_data * _mask[np.newaxis, ...]).sum(axis=flat_sig_dims)
        results.append(fill_nav_with_roi(ds_shape, masked_result, roi))
    naive_result = np.stack(results, axis=-1)

    return {'udf': udf, 'naive_result': {'intensity': naive_result}}


def fill_nav_with_roi(ds_shape, flat_result, roi, fill=np.nan):
    nav_shape = tuple(ds_shape.nav)
    if roi is None:
        return flat_result.reshape(nav_shape)
    else:
        full = np.full(nav_shape, fill_value=fill, dtype=flat_result.dtype).ravel()
        full[roi.ravel()] = flat_result
        return full.reshape(nav_shape)


def flatten_with_roi(data, roi, ds_shape):
    n_sig_dims = ds_shape.sig.dims
    flat_nav_data = data.reshape(-1, *tuple(ds_shape.sig))
    flat_sig_dims = tuple(range(-n_sig_dims, 0))
    if roi is not None:
        flat_nav_data = flat_nav_data[roi.ravel()]
    return flat_nav_data, flat_sig_dims


def build_udf_dict(udf_config, ds_dict):
    udf_dict = {'udf': [], 'naive_result': []}
    for udf in udf_config:
        factory = udf['factory']
        udf_params = udf['params']
        _udf_dict = factory(udf_params, ds_dict)

        udf_dict['udf'].append(_udf_dict['udf'])
        udf_dict['naive_result'].append(_udf_dict['naive_result'])

        if 'roi' in _udf_dict.keys():
            assert 'roi' not in udf_dict.keys(), 'Cannot define more than one custom ROI'
            udf_dict['roi'] = _udf_dict['roi']

    return udf_dict


def unpack_all(udf_results_dask):
    return tuple({k: v.data for k, v in udf_result.items()}
                 for udf_result in udf_results_dask)


def alldask(udf_results_dask):
    if not isinstance(udf_results_dask, tuple):
        udf_results_dask = (udf_results_dask,)
    for udf_result in udf_results_dask:
        for daskbuffer in udf_result.values():
            assert isinstance(daskbuffer.delayed_raw_data, da.Array)
            assert isinstance(daskbuffer.delayed_data, da.Array)
            # all dask, but these two:
            assert isinstance(daskbuffer.raw_data, np.ndarray)
            assert isinstance(daskbuffer.data, np.ndarray)


def try_convert_py(obj):
    try:
        return obj.item()
    except AttributeError:
        return obj


def allclose_with_nan(array1, array2, tol=None):
    if not isinstance(array1, np.ndarray) or not isinstance(array2, np.ndarray):
        assert try_convert_py(array1) == try_convert_py(array2)
        return
    assert array1.shape == array2.shape
    try:
        a1_nan = np.isnan(array1)
        a2_nan = np.isnan(array2)
        assert (a1_nan == a2_nan).all()
    except TypeError:
        # Cannot isnan non-numeric arrays
        pass
    tol_dict = {}
    if tol is not None:
        tol_dict = {'atol': tol}
    if np.issubdtype(array2.dtype, np.number):
        assert np.allclose(array1, array2, **tol_dict, equal_nan=True)
    else:
        assert (array1 == array2).all()


@pytest.mark.parametrize(
    "ds_config", ({'shape': (16, 8, 32, 64), 'tileshape': (8, 32, 64),
                   'num_partitions': 2, 'sig_dims': 2},
                  {'shape': (8, 16, 64, 64), 'tileshape': (8, 32, 64),
                   'num_partitions': 7, 'sig_dims': 2},
                  {'shape': (16, 16, 8, 64), 'tileshape': (8, 8, 64),
                   'num_partitions': 1, 'sig_dims': 2}),
    ids=['2part', '8part', '1part']
)
@pytest.mark.parametrize(
    "udf_config", (
        [
            {'factory': sigsum, 'params': {}}],
        [
            {'factory': navsum, 'params': {}}],
        [
            {'factory': numba_navsum, 'params': {}}],
        [
            {'factory': mask, 'params': {}}],
        [
            {'factory': count, 'params': {}}],
        [
            {'factory': string, 'params': {}}],
        [
            {'factory': mask, 'params': {'slices': [np.s_[:, :], np.s_[0:3, :]],
                                         'values': [1, 1.7]}}],
        [
            {'factory': sigsumdask, 'params': {}}],
        [
            {'factory': navsumdask, 'params': {}}],
        [
            {'factory': numba_navsumdask, 'params': {}}],
        [
            {'factory': navsum, 'params': {}},
            {'factory': sigsum, 'params': {}}],
        [
            {'factory': sigsum, 'params': {}},
            {'factory': navsumdask, 'params': {}},
            {'factory': numba_navsum, 'params': {}}],
        [
            {'factory': sigsumdask, 'params': {}},
            {'factory': navsum, 'params': {}},
            {'factory': numba_navsumdask, 'params': {}}],
                   ),
    ids=['sigsum', 'navsum', 'numba_navsum', 'mask',
         'count', 'string', 'dualmask',
         'sigsumdask', 'navsumdask', 'numba_navsumdask',
         'navsum+sigsum', 'sigsum+navsumdask+numba_navsum',
         'sigsumdask+navsum+numba_navsumdask']
)
@pytest.mark.parametrize(
    "use_roi", (True, False),
    ids=['w/ ROI', 'no ROI']
)
def test_udfs(delayed_ctx, ds_config, udf_config, use_roi):
    ds_dict = get_dataset(delayed_ctx, **ds_config, use_roi=use_roi)
    udf_dict = build_udf_dict(udf_config, ds_dict)
    if 'roi' in udf_dict.keys():
        ds_dict['roi'] = udf_dict['roi']

    dataset = ds_dict['dataset']
    udf = udf_dict['udf']
    roi_array = ds_dict['roi']
    if not use_roi and 'roi' not in udf_dict.keys():
        assert roi_array is None
    result_dask = delayed_ctx.run_udf(dataset=dataset, udf=udf, roi=roi_array)

    alldask(result_dask)
    computed_results = delayed_ctx.executor._compute(result_dask)
    for udf_computed_results, naive_results in zip(computed_results, udf_dict['naive_result']):
        for k, result in udf_computed_results.items():
            direct_result = naive_results[k]
            if direct_result is None:
                continue
            allclose_with_nan(result.data, direct_result, tol=udf_dict.get('tolerance', None))


def test_iterate(delayed_ctx, default_raw):
    udf = MySumUDF()
    res_noiter = delayed_ctx.run_udf(dataset=default_raw, udf=udf)
    for res in delayed_ctx.run_udf_iter(dataset=default_raw, udf=udf):
        print(res)
    # the end result is the same:
    assert np.allclose(res_noiter['intensity'], res.buffers[0]['intensity'])


@pytest.mark.asyncio
async def test_async(delayed_ctx, default_raw):
    udf = MySumUDF()
    res_async = await delayed_ctx.run_udf(dataset=default_raw, udf=udf, sync=False)
    res_sync = delayed_ctx.run_udf(dataset=default_raw, udf=udf, sync=True)
    assert np.allclose(res_async['intensity'], res_sync['intensity'])

    async for res in DelayedUDFRunner([udf]).run_for_dataset_async(
        default_raw, delayed_ctx.executor, roi=None, cancel_id="cancel"
    ):
        print(res)

    assert np.allclose(res_sync['intensity'], res.buffers[0]['intensity'])


def test_map(delayed_ctx):
    iterable = range(5)
    wrapped_delayed = delayed_ctx.executor.map(lambda x: x**2, iterable)
    assert np.allclose(wrapped_delayed, np.asarray(iterable)**2)


def test_bare_compute(delayed_ctx):
    ds_dict = get_dataset(delayed_ctx, (16, 8, 32, 32), (8, 32, 32), 4, 2)
    dataset = ds_dict['dataset']
    udf = MySumSigUDF()
    result_dask = delayed_ctx.run_udf(dataset=dataset, udf=udf)
    dask_res = result_dask['intensity'].delayed_data
    # Check one chunk per partition
    assert len(dask_res.chunks[0]) == 4
    res = dask.compute(dask_res)
    res = res[0]
    assert np.allclose(res, ds_dict['data'].sum(axis=(2, 3)))


def test_unwrap_null_case(delayed_ctx):
    nest = {'a': [5, 6, 7], 'b': (1, 2, {'x': 6, 'y': 'string'})}
    unwrapped = delayed_ctx.executor.unwrap_results(nest)
    for k, v in nest.items():
        assert isinstance(v, type(unwrapped[k]))
        for _i, _v in enumerate(v):
            assert isinstance(_v, type(unwrapped[k][_i]))
            if isinstance(_v, dict):
                for __k, __v in _v.items():
                    assert unwrapped[k][_i][__k] == __v
            else:
                assert unwrapped[k][_i] == _v


def test_only_dask(lt_ctx, delayed_ctx):
    ds_dict = get_dataset(delayed_ctx, (16, 8, 32, 32), (8, 32, 32), 4, 2)
    dataset = ds_dict['dataset']
    udf = OnlyDaskSumUDF()
    result_dask = delayed_ctx.run_udf(dataset=dataset, udf=udf)
    dask_res = result_dask['intensity'].delayed_data
    res = dask.compute(dask_res)
    res = res[0]
    assert np.allclose(res, ds_dict['data'].sum(axis=(0, 1)))

    with pytest.raises(NotImplementedError):
        lt_ctx.run_udf(dataset=dataset, udf=udf)


class BadSumMergeAllUDF(OnlyDaskSumUDF):
    def merge_all(self, ordered_results):
        intensity = np.stack([b.intensity for b in ordered_results.values()]).sum(axis=0)
        return {
            'intensity': intensity,
            'notaresult': intensity
        }


def test_bad_merge_all(delayed_ctx):
    ds_dict = get_dataset(delayed_ctx, (16, 8, 32, 32), (8, 32, 32), 4, 2)
    dataset = ds_dict['dataset']
    udf = BadSumMergeAllUDF()
    with pytest.raises(ValueError):
        delayed_ctx.run_udf(dataset=dataset, udf=udf)



import os
import stat
from pathlib import Path

import psutil

from libertem.io.utils import get_owner_name


class FSError(Exception):
    def __init__(self, msg, code, alternative=None):
        super().__init__(msg, code, alternative)
        self.code = code
        self.alternative = alternative


def _access_ok(path: Path):
    try:
        # May trigger PermissionError on Windows
        allowed = os.path.isdir(path) and os.access(path, os.R_OK | os.X_OK)
        if allowed:
            Path(path).resolve()
            # Windows sometimes raises an exception even if "allowed"
            os.listdir(path)
        return allowed
    except PermissionError:
        return False


def _get_alt_path(path: str):
    cur_path = Path(path).expanduser()
    try:
        cur_path = cur_path.resolve()
    # Triggered by empty DVD drive or permission denied on Windows
    except PermissionError:
        # we just skip resolving at this point and
        # move one up
        pass
    while not _access_ok(cur_path):
        try:
            cur_path = cur_path.parents[0]
            cur_path = cur_path.resolve()
        except IndexError:  # from cur_path.parents[0]
            # There is no parent
            return Path.home()
        except PermissionError:  # from cur_path.resolve()
            # No access yet, we move one up again
            pass
    return cur_path


def stat_path(path: str):
    try:
        return Path(path).expanduser().resolve().stat()
    except FileNotFoundError:
        raise FSError(
            code="NOT_FOUND",
            msg="path %s could not be found" % path,
            alternative=str(_get_alt_path(path)),
        )
    except PermissionError as e:
        raise FSError(
            code="PERMISSION_ERROR",
            msg=str(e),
            alternative=str(_get_alt_path(path)),
        )


def get_fs_listing(path: str):
    try:
        abspath = Path(path).expanduser().resolve()
    # Triggered by empty DVD drive on Windows
    except PermissionError as e:
        raise FSError(
            code="PERMISSION_ERROR",
            msg=str(e),
            alternative=str(_get_alt_path(path)),
        )
    if not abspath.is_dir():
        raise FSError(
            code="NOT_FOUND",
            msg="path %s could not be found" % path,
            alternative=str(_get_alt_path(path)),
        )
    if not _access_ok(abspath):
        raise FSError(
            code="ACCESS_DENIED",
            msg="access to %s was denied" % path,
            alternative=str(_get_alt_path(path)),
        )
    names = os.listdir(abspath)
    dirs = []
    files = []
    names = [".."] + names
    for name in names:
        full_path = os.path.join(abspath, name)
        try:
            s = os.stat(full_path)
        except (FileNotFoundError, PermissionError):
            # this can happen either because of a TOCTOU-like race condition
            # or for example for things like broken softlinks
            continue

        try:
            owner = get_owner_name(full_path, s)
        except FileNotFoundError:  # only from win_tweaks.py version
            continue
        except OSError:  # only from win_tweaks.py version
            owner = "<Unknown>"

        res = {"name": name, "stat": s, "owner": owner}
        if stat.S_ISDIR(s.st_mode):
            dirs.append(res)
        else:
            files.append(res)
    drives = [
        part.mountpoint
        for part in psutil.disk_partitions()
        if part.fstype != "squashfs"
    ]
    places = [
        {"key": "home", "title": "Home", "path": str(Path.home())},
    ]

    return {
        "path": str(abspath),
        "files": files,
        "dirs": dirs,
        "drives": drives,
        "places": places,
    }



import os
import contextlib
import typing
from typing import Optional
import warnings
import logging
import time

import numpy as np
import h5py
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common.math import prod, flat_nonzero
from libertem.common import Slice, Shape
from libertem.common.buffers import zeros_aligned
from libertem.io.corrections import CorrectionSet
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, Partition, DataTile, DataSetException, DataSetMeta,
    TilingScheme,
)


# alias for mocking:
current_time = time.time

logger = logging.getLogger(__name__)


class HDF5DatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/HDF5DatasetParams.schema.json",
        "title": "HDF5DatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "HDF5"},
            "path": {"type": "string"},
            "ds_path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
        },
        "required": ["type", "path", "ds_path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path", "ds_path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


def _ensure_2d_nav(nav_shape: tuple[int, ...]) -> tuple[int, int]:
    # For any iterable shape, reduce or pad it to a 2-tuple
    # with the same prod(shape). Reduction from left to right
    # (final dimension preserved). Special case for empty
    # nav_shape which is converted to (1, 1)
    nav_shape = tuple(nav_shape)
    if len(nav_shape) == 1:
        nav_shape = (1,) + nav_shape
    elif len(nav_shape) >= 2:
        nav_shape = (prod(nav_shape[:-1]),) + nav_shape[-1:]
    elif len(nav_shape) == 0:
        return (1, 1)
    else:
        raise ValueError(f'Incompatible nav_shape {nav_shape}')
    return nav_shape


class HDF5ArrayDescriptor(typing.NamedTuple):
    name: str
    shape: tuple[int, ...]
    dtype: np.dtype
    compression: Optional[str]
    chunks: tuple[int, ...]


def _get_datasets(path):
    datasets: list[HDF5ArrayDescriptor] = []

    try:
        timeout = int(os.environ.get('LIBERTEM_IO_HDF5_TIMEOUT_DEBUG', 3))
    except ValueError:
        timeout = 3

    t0 = current_time()

    def _make_list(name, obj):
        if current_time() - t0 > timeout:
            raise TimeoutError
        if hasattr(obj, 'size') and hasattr(obj, 'shape'):
            if obj.ndim < 3:
                # Can't process this dataset, skip
                return
            datasets.append(
                HDF5ArrayDescriptor(name, obj.shape, obj.dtype, obj.compression, obj.chunks)
            )

    with h5py.File(path, 'r') as f:
        f.visititems(_make_list)
    return datasets


def _have_contig_chunks(chunks, ds_shape):
    """
    Returns `True` if the `chunks` are contiguous in the navigation axes.

    Examples
    --------

    >>> ds_shape = Shape((64, 64, 128, 128), sig_dims=2)
    >>> _have_contig_chunks((1, 4, 32, 32), ds_shape)
    True
    >>> _have_contig_chunks((2, 4, 32, 32), ds_shape)
    False
    >>> _have_contig_chunks((2, 64, 32, 32), ds_shape)
    True
    >>> _have_contig_chunks((64, 1, 32, 32), ds_shape)
    False
    >>> ds_shape_5d = Shape((16, 64, 64, 128, 128), sig_dims=2)
    >>> _have_contig_chunks((1, 1, 2, 32, 32), ds_shape_5d)
    True
    >>> _have_contig_chunks((1, 2, 1, 32, 32), ds_shape_5d)
    False
    >>> _have_contig_chunks((2, 1, 1, 32, 32), ds_shape_5d)
    False
    """
    # In other terms:
    # There exists an index `i` such that `prod(chunks[:i]) == 1` and
    # chunks[i+1:] == ds_shape[i+1:], limited to the nav part of chunks and ds_shape
    #
    nav_shape = tuple(ds_shape.nav)
    nav_dims = len(nav_shape)
    chunks_nav = chunks[:nav_dims]

    for i in range(nav_dims):
        left = chunks_nav[:i]
        left_prod = prod(left)
        if left_prod == 1 and chunks_nav[i + 1:] == nav_shape[i + 1:]:
            return True
    return False


def _partition_shape_for_chunking(chunks, ds_shape):
    """
    Get the minimum partition shape for that allows us to prevent read amplification
    with chunked HDF5 files.

    Examples
    --------

    >>> ds_shape = Shape((64, 64, 128, 128), sig_dims=2)
    >>> _partition_shape_for_chunking((1, 4, 32, 32), ds_shape)
    (1, 4, 128, 128)
    >>> _partition_shape_for_chunking((2, 4, 32, 32), ds_shape)
    (2, 64, 128, 128)
    >>> _partition_shape_for_chunking((2, 64, 32, 32), ds_shape)
    (2, 64, 128, 128)
    >>> _partition_shape_for_chunking((64, 1, 32, 32), ds_shape)
    (64, 64, 128, 128)
    >>> ds_shape_5d = Shape((16, 64, 64, 128, 128), sig_dims=2)
    >>> _partition_shape_for_chunking((1, 1, 2, 32, 32), ds_shape_5d)
    (1, 1, 2, 128, 128)
    >>> _partition_shape_for_chunking((1, 2, 1, 32, 32), ds_shape_5d)
    (1, 2, 64, 128, 128)
    >>> _partition_shape_for_chunking((2, 1, 1, 32, 32), ds_shape_5d)
    (2, 64, 64, 128, 128)
    """
    first_non_one = [x == 1 for x in chunks].index(False)
    shape_left = chunks[:first_non_one + 1]
    return shape_left + ds_shape[first_non_one + 1:]


def _tileshape_for_chunking(chunks, ds_shape):
    """
    Calculate a tileshape for tiled reading from chunked
    data sets.

    Examples
    --------
    >>> ds_shape = Shape((64, 64, 128, 128), sig_dims=2)
    >>> _tileshape_for_chunking((1, 4, 32, 32), ds_shape)
    (4, 32, 32)
    """
    return chunks[-ds_shape.sig.dims - 1:]


def _get_tileshape_nd(partition_slice, tiling_scheme):
    extra_nav_dims = partition_slice.shape.nav.dims - tiling_scheme.shape.nav.dims
    # keep shape of the rightmost dimension:
    nav_item = min(tiling_scheme.shape[0], partition_slice.shape.nav[-1])
    return extra_nav_dims * (1,) + (nav_item,) + tuple(tiling_scheme.shape.sig)


class H5Reader:
    def __init__(self, path, ds_path):
        self._path = path
        self._ds_path = ds_path

    @contextlib.contextmanager
    def get_h5ds(self, cache_size=1024 * 1024):
        logger.debug("H5Reader.get_h5ds: cache_size=%dMiB", cache_size / 1024 / 1024)
        with h5py.File(self._path, 'r', rdcc_nbytes=cache_size, rdcc_nslots=19997) as f:
            yield f[self._ds_path]


class H5DataSet(DataSet):
    """
    Read data from a HDF5 data set.

    Examples
    --------

    >>> ds = ctx.load("hdf5", path=path_to_hdf5, ds_path="/data")

    Parameters
    ----------
    path: str
        Path to the file

    ds_path: str
        Path to the HDF5 data set inside the file

    nav_shape: tuple of int, optional
        A n-tuple that specifies the shape of the navigation / scan grid.
        By default this is inferred from the HDF5 dataset.

    sig_shape: tuple of int, optional
        A n-tuple that specifies the shape of the signal / frame grid.
        This parameter is currently unsupported and will raise an error
        if provided and not matching the underlying data sig shape.
        By default the sig_shape is inferred from the HDF5 dataset
        via the :code:`sig_dims` parameter.

    sig_dims: int
        Number of dimensions that should be considered part of the signal (for example
        2 when dealing with 2D image data)

    sync_offset: int, optional, by default 0
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    target_size: int
        Target partition size, in bytes. Usually doesn't need to be changed.

    min_num_partitions: int
        Minimum number of partitions, set to number of cores if not specified. Usually
        doesn't need to be specified.

    Note
    ----
    If the HDF5 file to be loaded contains compressed data
    using a custom compression filter (other than GZIP, LZF or SZIP),
    the associated HDF5 filter library must be imported on the workers
    before accessing the file. See the `h5py documentation
    on filter pipelines
    <https://docs.h5py.org/en/stable/high/dataset.html#filter-pipeline>`_
    for more information.

    The library `hdf5plugin <https://github.com/silx-kit/hdf5plugin>`_
    is preloaded automatically if it is installed. Other filter libraries
    may have to be specified for preloading by the user.

    Preloads for a local :class:`~libertem.executor.dask.DaskJobExecutor` can be
    specified through the :code:`preload` argument of either
    :meth:`~libertem.executor.dask.DaskJobExecutor.make_local` or
    :func:`libertem.executor.dask.cluster_spec`. For the
    :class:`libertem.executor.inline.InlineJobExecutor`, the plugins
    can simply be imported in the main script.

    For the web GUI or for running LiberTEM in a cluster with existing workers
    (e.g. by running
    :code:`libertem-worker` or :code:`dask-worker` on nodes),
    necessary imports can be specified as :code:`--preload` arguments to
    the launch command,
    for example with :code:`libertem-server --preload hdf5plugin` resp.
    :code:`libertem-worker --preload hdf5plugin tcp://scheduler_ip:port`.
    :code:`--preload` can be specified multiple times.
    """
    def __init__(self, path, ds_path=None, tileshape=None, nav_shape=None, sig_shape=None,
                 target_size=None, min_num_partitions=None, sig_dims=2, io_backend=None,
                 sync_offset: int = 0):
        super().__init__(io_backend=io_backend)
        if io_backend is not None:
            raise ValueError("H5DataSet currently doesn't support alternative I/O backends")
        self.path = path
        self.ds_path = ds_path
        self.target_size = target_size
        self.sig_dims = sig_dims
        # handle backwards-compatability:
        if tileshape is not None:
            warnings.warn(
                "tileshape argument is ignored and will be removed after 0.6.0",
                FutureWarning
            )
        # self.min_num_partitions appears to be never used
        self.min_num_partitions = min_num_partitions
        self._dtype = None
        self._shape = None
        self._nav_shape = nav_shape
        self._sig_shape = sig_shape
        self._sync_offset = sync_offset
        self._chunks = None
        self._compression = None

    def get_reader(self):
        return H5Reader(
            path=self.path,
            ds_path=self.ds_path
        )

    def _do_initialize(self):
        if self.ds_path is None:
            try:
                datasets = _get_datasets(self.path)
                largest_ds = max(datasets, key=lambda x: prod(x.shape))
            # FIXME: excepting `SystemError` temporarily
            # more info: https://github.com/h5py/h5py/issues/1740
            except (ValueError, TimeoutError, SystemError):
                raise DataSetException(f'Unable to infer dataset from file {self.path}')
            self.ds_path = largest_ds.name
        with self.get_reader().get_h5ds() as h5ds:
            self._dtype = h5ds.dtype
            shape = h5ds.shape
            if len(shape) == self.sig_dims:
                # shape = (1,) + shape -> this leads to indexing errors down the line
                # so we currently don't support opening 2D HDF5 files
                raise DataSetException("2D HDF5 files are currently not supported")
            ds_shape = Shape(shape, sig_dims=self.sig_dims)
            if self._sig_shape is not None and self._sig_shape != ds_shape.sig.to_tuple():
                raise DataSetException("sig reshaping currently not supported with HDF5 files")
            self._image_count = ds_shape.nav.size
            nav_shape = ds_shape.nav.to_tuple() if self._nav_shape is None else self._nav_shape
            self._shape = nav_shape + ds_shape.sig
            self._meta = DataSetMeta(
                shape=self.shape,
                raw_dtype=self._dtype,
                sync_offset=self._sync_offset,
                image_count=self._image_count,
                metadata={'ds_raw_shape': ds_shape}
            )
            self._chunks = h5ds.chunks
            self._compression = h5ds.compression
            if self._compression is not None:
                warnings.warn(
                    "Loading compressed HDF5, performance can be worse than with other formats",
                    RuntimeWarning
                )
            self._nav_shape_product = self._shape.nav.size
            self._sync_offset_info = self.get_sync_offset_info()
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    @classmethod
    def get_msg_converter(cls):
        return HDF5DatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"h5", "hdf5", "hspy", "nxs"}

    @classmethod
    def get_supported_io_backends(self):
        return []

    @classmethod
    def _do_detect(cls, path):
        try:
            with h5py.File(path, 'r'):
                pass
        except OSError as e:
            raise DataSetException(repr(e)) from e

    @classmethod
    def detect_params(cls, path, executor):
        try:
            executor.run_function(cls._do_detect, path)
        except (OSError, KeyError, ValueError, TypeError, DataSetException):
            # not a h5py file or can't open for some reason:
            return False

        # Read the dataset info from the file
        try:
            datasets = executor.run_function(_get_datasets, path)
            if not datasets:
                raise RuntimeError(f'Found no compatible datasets in the file {path}')
        # FIXME: excepting `SystemError` temporarily
        # more info: https://github.com/h5py/h5py/issues/1740
        except (RuntimeError, TimeoutError, SystemError):
            return {
                "parameters": {
                    "path": path,
                },
                "info": {
                    "datasets": [],
                }
            }

        # datasets contains at least one HDF5ArrayDescriptor
        # sig_dims is implicitly two here (for web GUI)
        sig_dims = 2
        full_info = [
            {
                "path": ds_item.name,
                "shape": ds_item.shape,
                "compression": ds_item.compression,
                "chunks": ds_item.chunks,
                "raw_nav_shape": ds_item.shape[:-sig_dims],
                "nav_shape": _ensure_2d_nav(ds_item.shape[:-sig_dims]),
                "sig_shape": ds_item.shape[-sig_dims:],
                "image_count": prod(ds_item.shape[:-sig_dims]),
            } for ds_item in datasets
        ]

        # use the largest size array as initial hdf5 dataset path
        # need to get info dict to access unpacked nav/sig shape
        # next line implements argmax on ds_descriptor.size
        ds_idx, _ = max(enumerate(datasets), key=lambda idx_x: prod(idx_x[1].shape))
        largest_ds = full_info[ds_idx]
        return {
            "parameters": {
                "path": path,
                "ds_path": largest_ds['path'],
                "nav_shape": largest_ds['nav_shape'],
                "sig_shape": largest_ds['sig_shape'],
            },
            "info": {
                "datasets": full_info
            }
        }

    @property
    def dtype(self):
        if self._dtype is None:
            raise RuntimeError("please call initialize")
        return self._dtype

    @property
    def shape(self):
        if self._shape is None:
            raise RuntimeError("please call initialize")
        return self._shape

    def check_valid(self):
        try:
            with self.get_reader().get_h5ds() as h5ds:
                h5ds.shape
            return True
        except (OSError, KeyError, ValueError) as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path": self.path,
            "ds_path": self.ds_path,
        }

    def get_diagnostics(self):
        with self.get_reader().get_h5ds() as ds:
            try:
                datasets = _get_datasets(self.path)
            # FIXME: excepting `SystemError` temporarily
            # more info: https://github.com/h5py/h5py/issues/1740
            except (TimeoutError, SystemError):
                datasets = []
            datasets = [
                {"name": descriptor.name, "value": [
                    {"name": "Size", "value": str(prod(descriptor.shape))},
                    {"name": "Shape", "value": str(descriptor.shape)},
                    {"name": "Datatype", "value": str(descriptor.dtype)},
                ]}
                for descriptor in sorted(
                    datasets, key=lambda i: prod(i.shape), reverse=True
                )
            ]
            return [
                {"name": "dtype", "value": str(ds.dtype)},
                {"name": "chunks", "value": str(ds.chunks)},
                {"name": "compression", "value": str(ds.compression)},
                {"name": "datasets", "value": datasets},
            ]

    def get_min_sig_size(self):
        if self._chunks is not None:
            return 1024  # allow for tiled processing w/ small-ish chunks
        # un-chunked HDF5 seems to prefer larger signal slices, so we aim for 32 4k blocks:
        return 32 * 4096 // np.dtype(self.meta.raw_dtype).itemsize

    def get_max_io_size(self) -> Optional[int]:
        if self._chunks is not None:
            # this may result in larger tile depth than necessary, but
            # it needs to be so big to pass the validation of the Negotiator. The tiles
            # won't ever be as large as the scheme dictates, anyway.
            # We limit it here to 256e6 elements, to also keep the chunk cache
            # usage reasonable:
            return int(256e6)
        return None  # use default value from Negotiator

    def get_base_shape(self, roi):
        if roi is not None:
            return (1,) + self.shape.sig
        if self._chunks is not None:
            sig_chunks = self._chunks[-self.shape.sig.dims:]
            return (1,) + sig_chunks
        return (1, 1,) + (self.shape[-1],)

    def adjust_tileshape(self, tileshape, roi):
        chunks = self._chunks
        sig_shape = self.shape.sig
        if roi is not None:
            return (1,) + sig_shape
        if chunks is not None and not _have_contig_chunks(chunks, self.shape):
            sig_chunks = chunks[-sig_shape.dims:]
            sig_ts = tileshape[-sig_shape.dims:]
            # if larger signal chunking is requested in the negotiation,
            # switch to full frames:
            if any(t > c for t, c in zip(sig_ts, sig_chunks)):
                # try to keep total tileshape size:
                tileshape_size = prod(tileshape)
                depth = max(1, tileshape_size // sig_shape.size)
                return (depth,) + sig_shape
            else:
                # depth needs to be limited to prod(chunks.nav)
                return _tileshape_for_chunking(chunks, self.shape)
        return tileshape

    def need_decode(self, roi, read_dtype, corrections):
        return True

    def get_partitions(self):
        # ds_shape = Shape(self.shape, sig_dims=self.sig_dims)
        ds_shape: Shape = self.meta['ds_raw_shape']
        ds_slice = Slice(origin=[0] * len(ds_shape), shape=ds_shape)
        target_size = self.target_size
        if target_size is None:
            if self._compression is None:
                target_size = 512 * 1024 * 1024
            else:
                target_size = 256 * 1024 * 1024
        partition_shape = self.partition_shape(
            target_size=target_size,
            dtype=self.dtype,
            containing_shape=ds_shape,
        ) + tuple(ds_shape.sig)

        # if the data is chunked in the navigation axes, choose a compatible
        # partition size (even important for non-compressed data!)
        chunks = self._chunks
        if chunks is not None and not _have_contig_chunks(chunks, ds_shape):
            partition_shape = _partition_shape_for_chunking(chunks, ds_shape)

        # -ve sync offset insert blank at beginning (skips at end)
        # +ve sync offset skips frames at beginning (blank at end)
        sync_offset = self._sync_offset
        ds_flat_shape = self.shape.flatten_nav()

        for slice_nd in ds_slice.subslices(partition_shape):
            raw_frames_slice = slice_nd.flatten_nav(ds_shape)
            raw_origin = raw_frames_slice.origin
            if self._sync_offset <= 0:
                # negative or zero s-o, shift right, clip length at end
                raw_frames_slice.origin = (raw_origin[0] + abs(sync_offset),) + raw_origin[1:]
            else:
                # positive s-o, shift left, clip part length at beginning
                corrected_nav_origin = raw_origin[0] - sync_offset
                if corrected_nav_origin < 0:
                    corrected_nav_size = raw_frames_slice.shape[0] + corrected_nav_origin
                    if corrected_nav_size <= 0:
                        # Empty partition, skip
                        continue
                    raw_frames_slice.shape = (corrected_nav_size,) + raw_frames_slice.shape.sig
                raw_frames_slice.origin = (max(0, corrected_nav_origin),) + raw_origin[1:]

            # All raw_frames_slice should have non-zero dims here
            partition_slice = raw_frames_slice.clip_to(ds_flat_shape)
            if any(v <= 0 for v in partition_slice.shape):
                # Empty partition after clip to desired shape, skip
                continue

            yield H5Partition(
                meta=self._meta,
                reader=self.get_reader(),
                partition_slice=partition_slice,
                slice_nd=slice_nd,
                io_backend=self.get_io_backend(),
                chunks=self._chunks,
                decoder=None,
                sync_offset=self._sync_offset,
            )

    def __repr__(self):
        return f"<H5DataSet of {self._dtype} shape={self._shape}>"


class H5Partition(Partition):
    def __init__(self, reader: H5Reader, slice_nd: Slice, chunks, sync_offset=0, *args, **kwargs):
        self.reader = reader
        self.slice_nd = slice_nd
        self._corrections = None
        self._chunks = chunks
        self._sync_offset = sync_offset
        super().__init__(*args, **kwargs)

    def _have_compatible_chunking(self):
        chunks = self._chunks
        if chunks is None:
            return True
        # all-1 in nav dims works:
        nav_dims = self.slice_nd.shape.nav.dims
        chunks_nav = self._chunks[:nav_dims]
        if all(c == 1 for c in chunks_nav):
            return True
        # everything else is problematic and needs special case:
        return False

    def _get_subslices_chunked_full_frame(self, scheme_lookup, nav_dims, tileshape_nd):
        """
        chunked full-frame reading. outer loop goes over the
        navigation coords of each chunk, inner loop is pushed into
        hdf5 by reading full frames need to order slices in a way that
        efficiently uses the chunk cache.
        """
        chunks_nav = self._chunks[:nav_dims]
        chunk_full_frame = chunks_nav + self.slice_nd.shape.sig
        chunk_slices = self.slice_nd.subslices(shape=chunk_full_frame)
        logger.debug(
            "_get_subslices_chunked_full_frame: chunking first by %r, then %r",
            chunk_full_frame, tileshape_nd,
        )
        for chunk_slice in chunk_slices:
            subslices = chunk_slice.subslices(shape=tileshape_nd)
            for subslice in subslices:
                idx = scheme_lookup[(subslice.origin[nav_dims:], subslice.shape[nav_dims:])]
                yield idx, subslice

    def _get_subslices_chunked_tiled(self, tiling_scheme, scheme_lookup, nav_dims, tileshape_nd):
        """
        general tiled reading w/ chunking outer loop is a chunk in
        signal dimensions, inner loop is over "rows in nav"
        """
        slice_nd_sig = self.slice_nd.sig
        slice_nd_nav = self.slice_nd.nav
        chunks_nav = self._chunks[:nav_dims]
        sig_slices = slice_nd_sig.subslices(tiling_scheme.shape.sig)
        logger.debug(
            "_get_subslices_chunked_tiled: chunking first by sig %r, then nav %r, finally %r",
            tiling_scheme.shape.sig, chunks_nav, tileshape_nd
        )
        for sig_slice in sig_slices:
            chunk_slices = slice_nd_nav.subslices(shape=chunks_nav)
            for chunk_slice_nav in chunk_slices:
                chunk_slice = Slice(
                    origin=chunk_slice_nav.origin + sig_slice.origin,
                    shape=chunk_slice_nav.shape + tuple(sig_slice.shape),
                )
                subslices = chunk_slice.subslices(shape=tileshape_nd)
                for subslice in subslices:
                    scheme_key = (subslice.origin[nav_dims:], subslice.shape[nav_dims:])
                    idx = scheme_lookup[scheme_key]
                    yield idx, subslice

    def _get_subslices(self, tiling_scheme):
        """
        Generate partition subslices for the given tiling scheme for the different cases.
        """
        if tiling_scheme.intent == "partition":
            tileshape_nd = self.slice_nd.shape
        else:
            tileshape_nd = _get_tileshape_nd(self.slice_nd, tiling_scheme)

        assert all(ts <= ps for (ts, ps) in zip(tileshape_nd, self.slice_nd.shape))

        nav_dims = self.slice_nd.shape.nav.dims

        # Three cases need to be handled:
        if self._have_compatible_chunking():
            # 1) no chunking, or compatible chunking. we are free to use
            #    whatever access pattern we deem efficient:
            logger.debug("using simple tileshape_nd slicing")
            subslices = self.slice_nd.subslices(shape=tileshape_nd)
            scheme_len = len(tiling_scheme)
            for idx, subslice in enumerate(subslices):
                scheme_idx = idx % scheme_len
                yield scheme_idx, subslice
        else:
            scheme_lookup = {
                (s.discard_nav().origin, tuple(s.discard_nav().shape)): idx
                for idx, s in tiling_scheme.slices
            }
            if len(tiling_scheme) == 1:
                logger.debug("using full-frame subslicing")
                yield from self._get_subslices_chunked_full_frame(
                    scheme_lookup, nav_dims, tileshape_nd
                )
            else:
                logger.debug("using chunk-adaptive subslicing")
                yield from self._get_subslices_chunked_tiled(
                    tiling_scheme, scheme_lookup, nav_dims, tileshape_nd
                )

    def _preprocess(self, tile_data, tile_slice):
        if self._corrections is None:
            return
        self._corrections.apply(tile_data, tile_slice)

    def _get_read_cache_size(self) -> float:
        chunks = self._chunks
        if chunks is None:
            return 1 * 1024 * 1024
        else:
            # heuristic on maximum chunk cache size based on number of cores
            # of the node this worker is running on, available memory, ...
            import psutil
            mem = psutil.virtual_memory()
            num_cores = psutil.cpu_count(logical=False)
            available: int = mem.available
            if num_cores is None:
                num_cores = 2
            cache_size: float = max(256 * 1024 * 1024, available * 0.8 / num_cores)
            return cache_size

    def _get_h5ds(self):
        cache_size = self._get_read_cache_size()
        return self.reader.get_h5ds(cache_size=cache_size)

    def _get_tiles_normal(self, tiling_scheme: TilingScheme, dest_dtype):
        with self._get_h5ds() as dataset:
            # because the dtype conversion done by HDF5 itself can be quite slow,
            # we need to use a buffer for reading in hdf5 native dtype:
            data_flat = np.zeros(tiling_scheme.shape, dtype=dataset.dtype).reshape((-1,))

            # ... and additionally a result buffer, for re-using the array used in the DataTile:
            data_flat_res = np.zeros(tiling_scheme.shape, dtype=dest_dtype).reshape((-1,))

            subslices = self._get_subslices(
                tiling_scheme=tiling_scheme,
            )

            sync_offset = self._sync_offset
            ds_num_frames = self.meta.shape.nav.size

            for scheme_idx, tile_slice in subslices:
                tile_slice_flat: Slice = tile_slice.flatten_nav(self.meta['ds_raw_shape'])
                raw_origin = tile_slice_flat.origin
                raw_shape = tile_slice_flat.shape

                # The following block translates from tile_slice in the raw array
                # to the partition coordinate system with sync_offset applied
                # By doing this before reading we can avoid reading some tiles
                # at the beginning/end of the dataset
                # We will still read tiles which partially overlap the nav space
                # and afterwards we drop the unecessary frames
                corrected_nav_origin = raw_origin[0] - sync_offset
                if corrected_nav_origin < 0:
                    # positive sync_offset, drop frames at beginning of DS
                    if abs(corrected_nav_origin) > tile_slice_flat.shape[0]:
                        # tile is completely before the first partition, can skip it
                        continue
                    # Clip at the beginning so adjust the tile shape
                    new_nav_size = tile_slice_flat.shape[0] + corrected_nav_origin
                    tile_slice_flat.shape = (new_nav_size,) + tile_slice_flat.shape.sig
                # Apply max(0, corrected_nav_origin) so we never provide negative nav coord
                tile_slice_flat.origin = (max(0, corrected_nav_origin),) + raw_origin[1:]
                # Now check for clipping at the end of the dataset
                final_frame_idx = tile_slice_flat.origin[0] + tile_slice_flat.shape[0]
                frames_beyond_end = final_frame_idx - ds_num_frames
                # We want to skip any tiles which are completely past the end of the dataset
                # and clip those which are only partially overlapping the final partition
                if frames_beyond_end >= tile_slice_flat.shape[0]:
                    # Empty tile after clip, skip
                    continue
                elif frames_beyond_end > 0:
                    # tile partially overlaps end of dataset, adjust the shape
                    new_nav_size = tile_slice_flat.shape[0] - frames_beyond_end
                    tile_slice_flat.shape = (new_nav_size,) + tile_slice_flat.shape.sig

                # Read the data in this block
                # cut buffer into the right size
                buf_size = tile_slice.shape.size
                buf = data_flat[:buf_size].reshape(tile_slice.shape)
                buf_res = data_flat_res[:buf_size].reshape(tile_slice.shape)
                dataset.read_direct(buf, source_sel=tile_slice.get())
                buf_res[:] = buf  # extra copy for faster dtype/endianess conversion
                tile_data = buf_res.reshape(raw_shape)

                # If the true tile origin is before the start of the dataset, must drop frames
                # This corresponds to the first raw tile which overlaps the first partition
                # and can only occur when sync_offset > 0
                if corrected_nav_origin < 0:
                    tile_data = tile_data[abs(corrected_nav_origin):, ...]

                # The final tiles in the dataset can partially overlap the final partition
                # Drop frames at end to match the partition size
                # we already verified if any frames will remain
                # this can occur for both +ve and -ve sync_offset
                if frames_beyond_end > 0:
                    tile_data = tile_data[:-frames_beyond_end, ...]

                # NOTE could the above two blocks ever apply simultaneously?
                # would the two operations conflict ?

                self._preprocess(tile_data, tile_slice_flat)
                yield DataTile(
                    tile_data,
                    tile_slice=tile_slice_flat,
                    scheme_idx=scheme_idx,
                )

    def _get_tiles_with_roi(self, roi, dest_dtype, tiling_scheme):
        # we currently don't chop up the frames when reading with a roi, so
        # the tiling scheme also must not contain more than one slice:
        # NOTE Why is this ??
        assert len(tiling_scheme) == 1, "incompatible tiling scheme! (%r)" % (tiling_scheme)

        flat_roi_nonzero = flat_nonzero(roi)
        start_at_frame = self.slice.origin[0]
        stop_at_frame = start_at_frame + self.slice.shape[0]
        part_mask = np.logical_and(flat_roi_nonzero >= start_at_frame,
                                   flat_roi_nonzero < stop_at_frame)
        # Must yield tiles with tile_slice in compressed nav dimension for roi
        frames_in_c_nav = np.arange(flat_roi_nonzero.size)[part_mask]
        frames_in_part = flat_roi_nonzero[part_mask]
        # -ve sync offset insert blank at beginning (skips at end)
        # +ve sync offset skips frames at beginning (blank at end)
        frames_in_raw = frames_in_part + self._sync_offset
        raw_shape = self.meta['ds_raw_shape'].nav.to_tuple()
        result_shape = Shape((1,) + tuple(self.meta.shape.sig), sig_dims=self.meta.shape.sig.dims)
        sig_origin = (0,) * self.meta.shape.sig.dims
        tile_data = np.zeros(result_shape, dtype=dest_dtype)

        with self._get_h5ds() as h5ds:
            tile_data_raw = np.zeros(result_shape, dtype=h5ds.dtype)
            for c_nav_idx, raw_idx in zip(frames_in_c_nav, frames_in_raw):
                tile_slice = Slice(
                    origin=(c_nav_idx,) + sig_origin,
                    shape=result_shape,
                )
                nav_coord = np.unravel_index(raw_idx, raw_shape)
                h5ds.read_direct(tile_data_raw, source_sel=nav_coord)
                tile_data[:] = tile_data_raw  # extra copy for dtype/endianess conversion
                self._preprocess(tile_data, tile_slice)
                yield DataTile(
                    tile_data,
                    tile_slice=tile_slice,
                    # there is only a single slice in the tiling scheme, so our
                    # scheme_idx is constant 0
                    scheme_idx=0,
                )

    def set_corrections(self, corrections: CorrectionSet):
        self._corrections = corrections

    def get_tiles(self, tiling_scheme: TilingScheme, dest_dtype="float32", roi=None,
            array_backend: Optional[ArrayBackend] = None):
        if array_backend is None:
            array_backend = self.meta.array_backends[0]
        assert array_backend in (NUMPY, CUDA)
        tiling_scheme = tiling_scheme.adjust_for_partition(self)
        if roi is not None:
            yield from self._get_tiles_with_roi(roi, dest_dtype, tiling_scheme)
        else:
            yield from self._get_tiles_normal(tiling_scheme, dest_dtype)

    def get_locations(self):
        return None

    def get_macrotile(self, dest_dtype="float32", roi=None):
        '''
        Return a single tile for the entire partition.

        This is useful to support process_partiton() in UDFs and to construct
        dask arrays from datasets.

        Note
        ----

        This can be inefficient if the dataset is compressed and chunked in the
        navigation axis, because you can either have forced-large macrotiles,
        or you can have read amplification effects, where a much larger amount
        of data is read from the HDF5 file than necessary.

        For example, if your chunking is :code:`(32, 32, 32, 32)`, in a
        dataset that is :code:`(128, 128, 256, 256)`, the partition must
        cover the whole of :code:`(32, 128, 256, 256)` - this is because
        partitions are contiguous in the navigation axis.

        The other possibility is to keep the partition smaller, for example
        only :code:`(3, 128, 256, 256)`. That would mean when reading a chunk
        from HDF5, we can only use 3*32 frames of the total 32*32 frames,
        a whopping ~10x read amplification.
        '''

        tileshape = self.shape
        if self._chunks is not None:
            tileshape = self._chunks

        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=Shape(tileshape, sig_dims=self.slice.shape.sig.dims).flatten_nav(),
            dataset_shape=self.meta.shape,
        )

        data = zeros_aligned(self.slice.adjust_for_roi(roi).shape, dtype=dest_dtype)

        for tile in self.get_tiles(
            tiling_scheme=tiling_scheme,
            dest_dtype=dest_dtype,
            roi=roi,
        ):
            rel_slice = tile.tile_slice.shift(self.slice)
            data[rel_slice.get()] = tile.data
        tile_slice = Slice(
            origin=(self.slice.origin[0], 0, 0),
            shape=Shape(data.shape, sig_dims=2),
        )
        return DataTile(
            data,
            tile_slice=tile_slice,
            scheme_idx=0,
        )



import numpy as np
import pytest

from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.memory import MemoryDataSet
from sparseconverter import (
    for_backend, get_device_class, SPARSE_BACKENDS, SPARSE_COO, NUMPY
)

from utils import _mk_random, set_device_class


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(SumSigUDF().get_backends())
)
def test_sum(lt_ctx, delayed_ctx, backend):
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random((16, 8, 32, 64), array_backend=SPARSE_COO)
        else:
            data = _mk_random((16, 8, 32, 64), array_backend=NUMPY)

        dataset = MemoryDataSet(
            data=data,
            tileshape=(8, 17, 23),
            num_partitions=2,
            sig_dims=2,
            array_backends=(backend, ) if backend is not None else None
        )

        sum_result = lt_ctx.run_udf(udf=SumSigUDF(), dataset=dataset)
        sum_delayed = delayed_ctx.run_udf(udf=SumSigUDF(), dataset=dataset)

        naive_result = for_backend(data.sum(axis=(2, 3)), NUMPY)

        assert np.allclose(sum_result['intensity'].data, naive_result)
        assert np.allclose(sum_delayed['intensity'].data, naive_result)



import numpy as np
import sparse
import pytest

from libertem.utils.generate import gradient_data, exclude_pixels
from libertem.io.corrections import detector
from libertem.common.sparse import is_sparse


REPEATS = 1


def _check_result(data, corrected, atol=1e-8, rtol=1e-5):
    m1 = np.unravel_index(np.argmax(np.abs(data - corrected)), data.shape)
    relative_error = np.zeros_like(data, dtype=np.result_type(data, np.float32))
    np.divide(
        np.abs(data - corrected),
        np.abs(data),
        out=relative_error,
        where=(data != 0)
    )
    m2 = np.unravel_index(
        np.argmax(relative_error), data.shape
    )
    # From np.allclose documentation
    select = np.abs(data - corrected) > (atol + rtol * np.abs(data))
    print(f"Maximum absolute error: data {data[m1]}, corrected {corrected[m1]}")
    print("Maximum relative error: data %s, corrected %s, relative %s" %
        (data[m2], corrected[m2], relative_error[m2])
    )
    print("Triggering tolerance limits: data %s, corrected %s, relative %s" %
        (data[select],
        corrected[select],
        relative_error[select])
    )

    assert np.allclose(data, corrected, atol=atol, rtol=rtol)


@pytest.mark.with_numba
def test_detector_correction():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([1, 2, 3])

        nav_dims = tuple(np.random.randint(low=1, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=1, high=16, size=num_sig_dims))

        data = gradient_data(nav_dims, sig_dims)

        # Test pure gain and offset correction without
        # patching pixels
        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=0)

        gain_map = np.random.random(sig_dims) + 1
        dark_image = np.random.random(sig_dims)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image

        assert np.allclose((damaged_data - dark_image) * gain_map, data)

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            dark_image=dark_image,
            gain_map=gain_map,
            excluded_pixels=exclude,
            inplace=False
        )

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )
        # Make sure we didn't do it in place
        assert not np.allclose(corrected, damaged_data)

        detector.correct(
            buffer=damaged_data,
            dark_image=dark_image,
            gain_map=gain_map,
            excluded_pixels=exclude,
            inplace=True
        )

        # Now damaged_data should be modified and equal to corrected
        # since it should have been done in place
        assert np.allclose(corrected, damaged_data)


@pytest.mark.with_numba
def test_detector_uint8():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([1, 2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        data = np.ones(nav_dims + sig_dims, dtype=np.uint8)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=2)

        gain_map = np.ones(sig_dims)
        dark_image = (np.random.random(sig_dims) * 3).astype(np.uint8)
        # Make sure the dark image is not all zero so that
        # the damaged data is different from the original
        # https://github.com/LiberTEM/LiberTEM/issues/910
        # This is only necessary for an integer dark image
        # since for float it would be extremely unlikely
        # that all values are exactly 0
        atleastone = np.random.randint(0, np.prod(sig_dims))
        dark_image[np.unravel_index(atleastone, sig_dims)] = 1

        damaged_data = data.copy()
        # We don't do that since it is set to 1 above
        # damaged_data /= gain_map
        damaged_data += dark_image

        damaged_data = damaged_data.astype(np.uint8)

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            dark_image=dark_image,
            gain_map=gain_map,
            excluded_pixels=exclude,
            inplace=False
        )

        assert corrected.dtype.kind == 'f'

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )
        # Make sure we didn't do it in place
        assert not np.allclose(corrected, damaged_data)

        with pytest.raises(TypeError):
            detector.correct(
                buffer=damaged_data,
                dark_image=dark_image,
                gain_map=gain_map,
                excluded_pixels=exclude,
                inplace=True
            )


@pytest.mark.with_numba
def test_detector_patch():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        data = gradient_data(nav_dims, sig_dims)

        gain_map = np.random.random(sig_dims) + 1
        dark_image = np.random.random(sig_dims)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=3)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            dark_image=dark_image,
            gain_map=gain_map,
            excluded_pixels=exclude,
            inplace=False
        )

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )


def test_detector_patch_large():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([2, 3])
        num_sig_dims = 2

        nav_dims = tuple(np.random.randint(low=3, high=5, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=4*32, high=1024, size=num_sig_dims))

        data = gradient_data(nav_dims, sig_dims)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=999)

        damaged_data = data.copy()
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            excluded_pixels=exclude,
            sig_shape=sig_dims,
            inplace=False
        )

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )


def test_detector_patch_too_large():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([2, 3])
        num_sig_dims = 2

        nav_dims = tuple(np.random.randint(low=3, high=5, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=4*32, high=1024, size=num_sig_dims))

        data = gradient_data(nav_dims, sig_dims)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=1001)

        damaged_data = data.copy()
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            excluded_pixels=exclude,
            sig_shape=sig_dims,
            inplace=False
        )

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
@pytest.mark.slow
def test_detector_patch_overlapping():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # Faithfully reconstruct in a constant dataset
        data = np.ones(nav_dims + sig_dims)

        gain_map = np.random.random(sig_dims) + 1
        dark_image = np.random.random(sig_dims)

        # Neighboring excluded pixels
        exclude = np.ones((num_sig_dims, 3), dtype=np.int32)
        exclude[0, 1] += 1
        exclude[1, 2] += 1

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        corrected = detector.correct(
            buffer=damaged_data,
            dark_image=dark_image,
            gain_map=gain_map,
            excluded_pixels=exclude,
            inplace=False
        )

        _check_result(
            data=data, corrected=corrected,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
def test_mask_correction():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # The mask-based correction is performed as float64 since it creates
        # numerical instabilities otherwise
        data = gradient_data(nav_dims, sig_dims).astype(np.float64)

        gain_map = (np.random.random(sig_dims) + 1).astype(np.float64)
        dark_image = np.random.random(sig_dims).astype(np.float64)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=0)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image

        assert np.allclose((damaged_data - dark_image) * gain_map, data)

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        masks = (np.random.random((2, ) + sig_dims) - 0.5).astype(np.float64)
        data_flat = data.reshape((np.prod(nav_dims), np.prod(sig_dims)))
        damaged_flat = damaged_data.reshape((np.prod(nav_dims), np.prod(sig_dims)))

        correct_dot = np.dot(data_flat, masks.reshape((-1, np.prod(sig_dims))).T)
        corrected_masks = detector.correct_dot_masks(masks, gain_map, exclude)

        assert not is_sparse(corrected_masks)

        reconstructed_dot =\
            np.dot(damaged_flat, corrected_masks.reshape((-1, np.prod(sig_dims))).T)\
            - np.dot(dark_image.flatten(), corrected_masks.reshape((-1, np.prod(sig_dims))).T)

        _check_result(
            data=correct_dot, corrected=reconstructed_dot,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
def test_mask_correction_sparse():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # The mask-based correction is performed as float64 since it creates
        # numerical instabilities otherwise
        data = gradient_data(nav_dims, sig_dims).astype(np.float64)

        gain_map = (np.random.random(sig_dims) + 1).astype(np.float64)
        dark_image = np.random.random(sig_dims).astype(np.float64)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=0)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        masks = sparse.DOK(sparse.zeros((20, ) + sig_dims, dtype=np.float64))
        indices = [np.random.randint(low=0, high=s, size=s//2) for s in (20, ) + sig_dims]
        for tup in zip(*indices):
            masks[tup] = 1
        masks = masks.to_coo()

        data_flat = data.reshape((np.prod(nav_dims), np.prod(sig_dims)))
        damaged_flat = damaged_data.reshape((np.prod(nav_dims), np.prod(sig_dims)))

        correct_dot = sparse.dot(data_flat, masks.reshape((-1, np.prod(sig_dims))).T)
        corrected_masks = detector.correct_dot_masks(masks, gain_map, exclude)
        assert is_sparse(corrected_masks)

        reconstructed_dot =\
            sparse.dot(damaged_flat, corrected_masks.reshape((-1, np.prod(sig_dims))).T)\
            - sparse.dot(dark_image.flatten(), corrected_masks.reshape((-1, np.prod(sig_dims))).T)

        _check_result(
            data=correct_dot, corrected=reconstructed_dot,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
def test_mask_patch():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # The mask-basedcorrection is performed as float64 since it creates
        # numerical instabilities otherwise
        data = gradient_data(nav_dims, sig_dims).astype(np.float64)

        gain_map = (np.random.random(sig_dims) + 1).astype(np.float64)
        dark_image = np.random.random(sig_dims).astype(np.float64)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=3)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        masks = (np.random.random((2, ) + sig_dims) - 0.3).astype(np.float64)
        data_flat = data.reshape((np.prod(nav_dims), np.prod(sig_dims)))
        damaged_flat = damaged_data.reshape((np.prod(nav_dims), np.prod(sig_dims)))

        correct_dot = np.dot(data_flat, masks.reshape((-1, np.prod(sig_dims))).T)
        corrected_masks = detector.correct_dot_masks(masks, gain_map, exclude)

        assert not is_sparse(corrected_masks)

        reconstructed_dot =\
            np.dot(damaged_flat, corrected_masks.reshape((-1, np.prod(sig_dims))).T)\
            - np.dot(dark_image.flatten(), corrected_masks.reshape((-1, np.prod(sig_dims))).T)

        _check_result(
            data=correct_dot, corrected=reconstructed_dot,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
def test_mask_patch_sparse():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # The mask-based correction is performed as float64 since it creates
        # numerical instabilities otherwise
        data = gradient_data(nav_dims, sig_dims).astype(np.float64)

        gain_map = (np.random.random(sig_dims) + 1).astype(np.float64)
        dark_image = np.random.random(sig_dims).astype(np.float64)

        exclude = exclude_pixels(sig_dims=sig_dims, num_excluded=3)

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        masks = sparse.DOK(sparse.zeros((20, ) + sig_dims, dtype=np.float64))
        indices = [np.random.randint(low=0, high=s, size=s//2) for s in (20, ) + sig_dims]
        for tup in zip(*indices):
            masks[tup] = 1
        masks = masks.to_coo()

        data_flat = data.reshape((np.prod(nav_dims), np.prod(sig_dims)))
        damaged_flat = damaged_data.reshape((np.prod(nav_dims), np.prod(sig_dims)))

        correct_dot = sparse.dot(data_flat, masks.reshape((-1, np.prod(sig_dims))).T)
        corrected_masks = detector.correct_dot_masks(masks, gain_map, exclude)
        assert is_sparse(corrected_masks)

        reconstructed_dot =\
            sparse.dot(damaged_flat, corrected_masks.reshape((-1, np.prod(sig_dims))).T)\
            - sparse.dot(dark_image.flatten(), corrected_masks.reshape((-1, np.prod(sig_dims))).T)

        _check_result(
            data=correct_dot, corrected=reconstructed_dot,
            atol=1e-8, rtol=1e-5
        )


@pytest.mark.with_numba
def test_mask_patch_overlapping():
    for i in range(REPEATS):
        print(f"Loop number {i}")
        num_nav_dims = np.random.choice([1, 2, 3])
        num_sig_dims = np.random.choice([2, 3])

        nav_dims = tuple(np.random.randint(low=8, high=16, size=num_nav_dims))
        sig_dims = tuple(np.random.randint(low=8, high=16, size=num_sig_dims))

        # The mask-based correction is performed as float64 since it creates
        # numerical instabilities otherwise
        # Constant data to reconstruct neighboring damaged pixels faithfully
        data = np.ones(nav_dims + sig_dims, dtype=np.float64)

        gain_map = (np.random.random(sig_dims) + 1).astype(np.float64)
        dark_image = np.random.random(sig_dims).astype(np.float64)

        # Neighboring excluded pixels
        exclude = np.ones((num_sig_dims, 3), dtype=np.int32)
        exclude[0, 1] += 1
        exclude[1, 2] += 1

        damaged_data = data.copy()
        damaged_data /= gain_map
        damaged_data += dark_image
        damaged_data[(Ellipsis, *exclude)] = 1e24

        print("Nav dims: ", nav_dims)
        print("Sig dims:", sig_dims)
        print("Exclude: ", exclude)

        masks = (np.random.random((2, ) + sig_dims) - 0.5).astype(np.float64)
        data_flat = data.reshape((np.prod(nav_dims), np.prod(sig_dims)))
        damaged_flat = damaged_data.reshape((np.prod(nav_dims), np.prod(sig_dims)))

        correct_dot = np.dot(data_flat, masks.reshape((-1, np.prod(sig_dims))).T)
        corrected_masks = detector.correct_dot_masks(masks, gain_map, exclude)

        assert not is_sparse(corrected_masks)

        reconstructed_dot =\
            np.dot(damaged_flat, corrected_masks.reshape((-1, np.prod(sig_dims))).T)\
            - np.dot(dark_image.flatten(), corrected_masks.reshape((-1, np.prod(sig_dims))).T)

        _check_result(
            data=correct_dot, corrected=reconstructed_dot,
            atol=1e-8, rtol=1e-5
        )



import os
import io

import numpy as np
import numba
from numba.typed import Dict
import contextlib
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common.math import prod
from libertem.io.dataset.base.backend import IOBackend, IOBackendImpl
from libertem.io.dataset.base.fileset import FileSet
from libertem.common import Shape, Slice
from libertem.common.buffers import BufferPool, ManagedBuffer
from libertem.common.numba import cached_njit
from .tiling import DataTile
from .decode import DtypeConversionDecoder

_r_n_d_cache = {}


def _make_buffered_reader_and_decoder(decode):
    """
    decode: from buffers, in bytes, possibly interpreted as native_dtype, to out_decoded.dtype
    """
    @cached_njit(boundscheck=False, nogil=True)
    def _buffered_tilereader(outer_idx, buffers, sig_dims, tile_read_ranges,
                           out_decoded, native_dtype, do_zero,
                           origin, shape, ds_shape, offsets):
        if do_zero:
            out_decoded[:] = 0
        for rr_idx in range(tile_read_ranges.shape[0]):
            rr = tile_read_ranges[rr_idx]
            if rr[1] == rr[2] == 0:
                break
            buf = buffers[rr[0]]
            offset = offsets[rr[0]]  # per-file offset of start of read buffer
            decode(
                inp=buf[rr[1] - offset:rr[2] - offset],
                out=out_decoded,
                idx=rr_idx,
                native_dtype=native_dtype,
                rr=rr,
                origin=origin,
                shape=shape,
                ds_shape=ds_shape,
            )
        return out_decoded
    return _buffered_tilereader


@numba.njit(cache=True, nogil=True)
def block_get_min_fill_factor(rrs):
    """
    Try to find out how sparse the given read ranges are, per file.

    Returns the smallest fill factor and maximum required buffer size.
    """
    # FIXME: replace with fixed arrays? If this fn stands out on a profile...
    min_per_file = {}
    max_per_file = {}
    payload_bytes = {}
    for tile_idx in range(rrs.shape[0]):
        for part_idx in range(rrs.shape[1]):
            part_rr = rrs[tile_idx, part_idx]
            fileno = part_rr[0]

            if part_rr[2] == part_rr[1] == 0:
                continue

            if fileno not in min_per_file:
                min_per_file[fileno] = part_rr[1]
            else:
                min_per_file[fileno] = min(part_rr[1], min_per_file[fileno])

            if fileno not in max_per_file:
                max_per_file[fileno] = part_rr[2]
            else:
                max_per_file[fileno] = max(part_rr[2], max_per_file[fileno])

            if fileno not in payload_bytes:
                payload_bytes[fileno] = 0
            payload_bytes[fileno] += part_rr[2] - part_rr[1]

    min_fill_factor = -1
    max_buffer_size = 0

    for k in min_per_file:
        read_size = max_per_file[k] - min_per_file[k]
        fill_factor = payload_bytes[k] / (read_size)
        # print(k, fill_factor, max_per_file[k], min_per_file[k], payload_bytes[k])
        if fill_factor < min_fill_factor or min_fill_factor == -1:
            min_fill_factor = fill_factor
        if read_size > max_buffer_size:
            max_buffer_size = read_size
    return min_fill_factor, max_buffer_size, min_per_file, max_per_file


class BufferedFile:
    def __init__(self, path, desc):
        self.path = path
        self.desc = desc
        self._handle = None
        self._arr = None

    def get_blocksize(self):
        """
        Return the block size to which reads should be aligned to. In case of normal
        buffered I/O, this can be 1
        """
        return 1

    def open(self):
        # disable internal I/O buffering, as we want to read directly
        # into our own buffer here:
        self._handle = open(self.path, "rb", buffering=0)
        return self

    def close(self):
        self._handle.close()
        self._handle = None

    @property
    def handle(self):
        if self._handle is None:
            raise RuntimeError("trying to access file handle of closed file")
        return self._handle

    def seek(self, offset):
        self._handle.seek(offset)

    def readinto(self, buf):
        BLOCKSIZE = self.get_blocksize()
        buf_orig = buf
        buf = memoryview(buf)
        to_read = len(buf)
        offset = 0
        last_to_read = to_read
        # `readinto` may return early, so we may need to re-run it:
        while to_read > 0:
            # Make sure reads are aligned to the blocksize,
            # after an early return of `readinto` to allow Direct I/O on Windows.
            # On Linux, a misaligned read which turns out to be at the end of the
            # file doesn't cause an error
            blockcount, remainder = divmod(offset, BLOCKSIZE)
            to_read += remainder
            offset = blockcount * BLOCKSIZE
            if remainder > 0:
                self._handle.seek(-remainder, io.SEEK_CUR)
            bytes_read = self._handle.readinto(buf[offset:])
            if bytes_read == 0:
                break
            to_read -= bytes_read
            offset += bytes_read
            # The block aligning code would otherwise
            # try to read a "tail" in an infinite loop if the file size
            # is not a multiple of BLOCKSIZE
            if to_read == last_to_read:
                break
            last_to_read = to_read
        return buf_orig[:offset]


class DirectBufferedFile(BufferedFile):
    def get_blocksize(self):
        """
        Return the block size to which reads should be aligned to. On Linux,
        this should be a multiple of the logical block size of the file system.
        """
        # NOTE: if required, this can be replaced with the appropriate syscall
        # to determine the block size. On Linux, this would be BLKBSZGET.
        return 4096

    def open(self):
        # disable internal I/O buffering, as we want to read directly
        # into our own buffer here:
        if hasattr(os, 'O_DIRECT'):
            self._fh = os.open(self.path, os.O_RDONLY | os.O_DIRECT)
            self._handle = open(self._fh, "rb", buffering=0)
        elif os.name == 'nt':
            import win32file
            import msvcrt
            # Make sure to keep a reference, otherwise
            # it will be closed
            self._fh = win32file.CreateFile(
                self.path,  # fileName
                win32file.GENERIC_READ,  # desiredAccess
                win32file.FILE_SHARE_READ
                | win32file.FILE_SHARE_WRITE
                | win32file.FILE_SHARE_DELETE,  # shareMode
                None,  # attributes
                win32file.OPEN_EXISTING,  # CreationDisposition
                # O_DIRECT equivalent (?)
                win32file.FILE_FLAG_NO_BUFFERING,  # flagsAndAttributes
                0,  # hTemplateFile
            )
            fd = msvcrt.open_osfhandle(int(self._fh), os.O_RDONLY)
            self._handle = os.fdopen(fd, "rb", buffering=0)
        else:
            raise RuntimeError("Direct I/O not supported on this platform.")
        return self

    def close(self):
        super().close()
        self._fh = None


class BufferedBackend(IOBackend, id_="buffered"):
    """
    I/O backend using a buffered reading strategy. Useful for slower media
    like HDDs, where seeks cause performance drops. Used by default
    on Windows.

    This does not perform optimally on SSDs under all circumstances, for
    better best-case performance, try using
    :class:`~libertem.io.dataset.base.MMapBackend` instead.

    Parameters
    ----------
    max_buffer_size : int
        Maximum buffer size, in bytes. This is passed to the tileshape
        negotiation to select the right depth.
    """
    def __init__(self, max_buffer_size=16*1024*1024):
        self._max_buffer_size = max_buffer_size

    @classmethod
    def from_json(cls, msg):
        """
        Construct an instance from the already-decoded `msg`.
        """
        raise NotImplementedError("TODO! implement me!")

    def get_impl(self):
        return BufferedBackendImpl(
            max_buffer_size=self._max_buffer_size,
        )


class BufferedBackendImpl(IOBackendImpl):
    def __init__(self, max_buffer_size, direct_io=False):
        super().__init__()
        self._max_buffer_size = max_buffer_size
        self._direct_io = direct_io
        self._buffer_pool = BufferPool()

    @contextlib.contextmanager
    def open_files(self, fileset: FileSet):
        cls: type[BufferedFile]
        if self._direct_io:
            cls = DirectBufferedFile
        else:
            cls = BufferedFile
        files = [
            cls(path=f.path, desc=f).open()
            for f in fileset
        ]
        yield files
        for f in files:
            f.close()

    def need_copy(
        self, decoder, roi, native_dtype, read_dtype, tiling_scheme=None, fileset=None,
        sync_offset=0, corrections=None,
    ):
        return True  # we always copy in this backend

    def get_read_and_decode(self, decode):
        key = (decode, "read")
        if key in _r_n_d_cache:
            return _r_n_d_cache[key]
        r_n_d = _make_buffered_reader_and_decoder(decode=decode)
        _r_n_d_cache[key] = r_n_d
        return r_n_d

    def get_max_io_size(self):
        return self._max_buffer_size

    def _get_tiles_by_block(
        self, tiling_scheme, open_files, read_ranges, read_dtype, native_dtype, decoder=None,
        corrections=None, sync_offset=0,
    ):
        if decoder is None:
            decoder = DtypeConversionDecoder()
        decode = decoder.get_decode(
            native_dtype=np.dtype(native_dtype),
            read_dtype=np.dtype(read_dtype),
        )
        r_n_d = self._r_n_d = self.get_read_and_decode(decode)

        native_dtype = decoder.get_native_dtype(native_dtype, read_dtype)

        sig_dims = tiling_scheme.shape.sig.dims
        ds_shape = np.array(tiling_scheme.dataset_shape)

        largest_slice = sorted((
            (prod(s_.shape), s_)
            for _, s_ in tiling_scheme.slices
        ), key=lambda x: x[0], reverse=True)[0][1]

        buf_shape = (tiling_scheme.depth,) + tuple(largest_slice.shape)

        need_clear = decoder.do_clear()

        slices = read_ranges[0]
        # Use NumPy prod for multidimensional array and axis parameter
        shape_prods = np.prod(slices[..., 1, :], axis=1, dtype=np.int64)
        ranges = read_ranges[1]
        scheme_indices = read_ranges[2]
        tile_block_size = len(tiling_scheme)

        with self._buffer_pool.empty(buf_shape, dtype=read_dtype) as out_decoded:
            out_decoded = out_decoded.reshape((-1,))
            for block_idx in range(0, slices.shape[0], tile_block_size):
                block_ranges = ranges[block_idx:block_idx + tile_block_size]

                fill_factor, req_buf_size, min_per_file, max_per_file = block_get_min_fill_factor(
                    block_ranges
                )
                # TODO: if it makes sense, implement sparse variant
                # if req_buf_size > self._max_buffer_size or fill_factor < self._sparse_threshold:
                yield from self._read_block_dense(
                    block_idx, tile_block_size, min_per_file, max_per_file, open_files,
                    slices, ranges, scheme_indices, shape_prods, out_decoded, r_n_d,
                    sig_dims, ds_shape, need_clear, native_dtype, corrections,
                )

    def _read_block_dense(
        self, block_idx, tile_block_size, min_per_file, max_per_file, open_files,
        slices, ranges, scheme_indices, shape_prods, out_decoded, r_n_d,
        sig_dims, ds_shape, need_clear, native_dtype, corrections,
    ):
        """
        Reads a block of tiles, starting at `block_idx`, having a size of
        `tile_block_size` read range entries.
        """
        # phase 1: read
        buffers = Dict()

        # this list manages the lifetime of the ManagedBuffer instances;
        # after `buf_ref` goes out of scope, the buffers are returned to
        # the buffer pool, so make sure that this matches with the usage
        # of the buffers!
        buf_ref = []
        for fileno in min_per_file.keys():
            fh = open_files[fileno]
            # add align_to to allow for alignment cut:
            align_to = fh.get_blocksize()
            read_size = max_per_file[fileno] - min_per_file[fileno] + align_to
            # ManagedBuffer gives us memory in 4k blocks, so the size is 4k aligned
            mb = ManagedBuffer(self._buffer_pool, read_size, alignment=fh.get_blocksize())
            arr = np.frombuffer(mb.buf, dtype=np.uint8)
            buf_ref.append(mb)
            seek_pos = min_per_file[fileno]
            alignment = 0
            # seek_pos needs to be aligned to 4k block size, too:
            if seek_pos % align_to != 0:
                alignment = seek_pos % align_to
                seek_pos = align_to * (seek_pos // align_to)
            fh.seek(seek_pos)
            read_result = fh.readinto(arr)
            # read may be truncated, if the buffer is larger than the file; we
            # truncate the buffer, too, to make sure we don't use any
            # uninitialized values. Also cut off `alignment` bytes at the beginning,
            # which were read to make O_DIRECT happy:
            buffers[fileno] = read_result[alignment:]

        # phase 2: decode tiles from the data that was read
        for idx in range(block_idx, block_idx + tile_block_size):
            origin = slices[idx, 0]
            shape = slices[idx, 1]
            tile_ranges = ranges[idx]
            scheme_idx = scheme_indices[idx]
            out_cut = out_decoded[:shape_prods[idx]].reshape((shape[0], -1))

            data = r_n_d(
                idx,
                buffers, sig_dims, tile_ranges,
                out_cut, native_dtype, do_zero=need_clear,
                origin=origin, shape=shape, ds_shape=ds_shape,
                offsets=min_per_file,
            )
            tile_slice = Slice(
                origin=origin,
                shape=Shape(shape, sig_dims=sig_dims)
            )
            data = data.reshape(shape)
            self.preprocess(data, tile_slice, corrections)
            yield DataTile(
                data,
                tile_slice=tile_slice,
                scheme_idx=scheme_idx,
            )

    def get_tiles(
        self, tiling_scheme, fileset, read_ranges, roi, native_dtype, read_dtype, decoder,
        sync_offset, corrections, array_backend: ArrayBackend,
    ):
        assert array_backend in (NUMPY, CUDA)
        with self.open_files(fileset) as open_files:
            yield from self._get_tiles_by_block(
                tiling_scheme=tiling_scheme,
                open_files=open_files,
                read_ranges=read_ranges,
                read_dtype=read_dtype,
                native_dtype=native_dtype,
                decoder=decoder,
                corrections=corrections,
                sync_offset=sync_offset,
            )



import os
import shutil

import pytest

import utils


def _raw_on_workers(tmpdirpath, datadir, filename, ctx):
    data = utils._mk_random(size=(16, 16, 128, 128), dtype='float32')

    def _make_example_raw():
        # workers don't automatically have the pytest tmp directory, create it:
        if not os.path.exists(tmpdirpath):
            os.makedirs(tmpdirpath)
        print("creating %s" % filename)
        data.tofile(filename)
        print("created %s" % filename)
        return tmpdirpath, os.listdir(tmpdirpath)

    import cloudpickle
    import pickle
    dumped = cloudpickle.dumps(_make_example_raw)

    pickle.loads(dumped)

    print("raw_on_workers _make_example_raw: %s" %
          (ctx.executor.run_each_host(_make_example_raw),))

    ds = ctx.load("raw",
                  path=str(filename),
                  nav_shape=(16, 16),
                  sig_shape=(128, 128),
                  dtype="float32")
    return ds


@pytest.fixture
def raw_on_workers(dist_ctx, tmpdir_factory):
    """
    copy raw dataset to each worker
    """

    datadir = tmpdir_factory.mktemp('data')
    filename = str(datadir + '/raw-test-on-workers')
    tmpdirpath = os.path.dirname(filename)

    ds = _raw_on_workers(tmpdirpath, datadir, filename, dist_ctx)
    yield ds

    def _cleanup():
        # NOTE: can't call a function defined in conftest here, as `conftest` is not available
        # as a module on the worker nodes
        files = os.listdir(tmpdirpath)
        shutil.rmtree(tmpdirpath, ignore_errors=True)
        print("removed %s" % tmpdirpath)
        return tmpdirpath, files

    print(f"raw_on_workers cleanup: {dist_ctx.executor.run_each_host(_cleanup)}")


@pytest.fixture
def raw_on_workers_ipy(ipy_ctx, tmpdir_factory):
    """
    copy raw dataset to each worker
    """

    datadir = tmpdir_factory.mktemp('data')
    filename = str(datadir + '/raw-test-on-workers')
    tmpdirpath = os.path.dirname(filename)

    ds = _raw_on_workers(tmpdirpath, datadir, filename, ipy_ctx)
    yield ds

    def _cleanup():
        # NOTE: can't call a function defined in conftest here, as `conftest` is not available
        # as a module on the worker nodes
        files = os.listdir(tmpdirpath)
        shutil.rmtree(tmpdirpath, ignore_errors=True)
        print("removed %s" % tmpdirpath)
        return tmpdirpath, files

    print(f"raw_on_workers cleanup: {ipy_ctx.executor.run_each_host(_cleanup)}")



import logging

import tornado.web
import tornado.gen
import tornado.websocket
import tornado.ioloop
import tornado.escape

from .base import log_message
from .messages import Message
from .state import SharedState

log = logging.getLogger(__name__)


class ConfigHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def get(self):
        log.info("ConfigHandler.get")
        msg = Message().config(config=self.state.get_config())
        log_message(msg)
        self.write(msg)


class ClusterDetailHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def get(self):
        details = await self.state.executor_state.get_resource_details()
        msg = Message().cluster_details(details=details)
        log_message(msg)
        self.write(msg)



import asyncio
import queue
import logging
import concurrent.futures

log = logging.getLogger(__name__)


class MessagePump:
    def __init__(self, event_bus, event_registry):
        self._event_bus = event_bus
        self._event_registry = event_registry

    async def run(self):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            loop = asyncio.get_running_loop()
            while True:
                try:
                    msg = await loop.run_in_executor(pool, self._event_bus.get)
                    log.debug("MessagePump.run: got msg: %s", msg)
                    await self._event_registry.broadcast_event(msg)
                except queue.Empty:
                    pass


class EventBus:
    """
    An internal event bus used in the web API for forwarding messages from any
    thread to all websocket connections.

    Do not use for latecy-sensitive messages.
    """

    def __init__(self):
        # unlimited size so we can put items without waiting
        self._queue = queue.Queue()

    def send(self, msg):
        log.debug("EventBus.send: got msg: %s", msg)
        self._queue.put(msg)

    def get(self, timeout=1.0):
        # Use a non-blocking `get` here, so we don't completely block the
        # thread we are running in. raises `queue.Empty` if there's nothing to
        # get.
        return self._queue.get(block=False, timeout=timeout)



import numpy as np

from libertem.analysis.sumsig import SumSigAnalysis
from libertem.io.dataset.memory import MemoryDataSet


def test_sumsig_analysis_smoke(lt_ctx):
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    analysis = SumSigAnalysis(dataset=dataset, parameters={})
    lt_ctx.run(analysis)



import os
import pathlib
import sys
import json
from unittest import mock
import random
import warnings

import numpy as np
import pytest
import cloudpickle

from libertem.io.dataset.mib import MIBDataSet, get_filenames
from libertem.udf.raw import PickUDF
from libertem.analysis.raw import PickFrameAnalysis
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.base import (
    TilingScheme, BufferedBackend, MMapBackend, DirectBackend
)

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

try:
    import rsciio
except ModuleNotFoundError:
    rsciio = None


MIB_TESTDATA_PATH = os.path.join(get_testdata_path(), 'default.mib')
HAVE_MIB_TESTDATA = os.path.exists(MIB_TESTDATA_PATH)

needsdata = pytest.mark.skipif(not HAVE_MIB_TESTDATA, reason="need .mib testdata")  # NOQA


@pytest.fixture
def default_mib(lt_ctx):
    nav_shape = (32, 32)
    ds = lt_ctx.load(
        "mib",
        path=MIB_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(4)
    return ds


@pytest.fixture
def default_mib_readahead(lt_ctx):
    nav_shape = (32, 32)
    ds = lt_ctx.load(
        "mib",
        path=MIB_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=MMapBackend(enable_readahead_hints=True),
    )
    ds.set_num_cores(4)
    return ds


@pytest.fixture
def buffered_mib(lt_ctx):
    buffered = BufferedBackend()
    ds = lt_ctx.load(
        "mib",
        path=MIB_TESTDATA_PATH,
        nav_shape=(32, 32),
        io_backend=buffered,
    )
    return ds


@pytest.fixture
def direct_mib(lt_ctx):
    direct = DirectBackend()
    ds = lt_ctx.load(
        "mib",
        path=MIB_TESTDATA_PATH,
        nav_shape=(32, 32),
        io_backend=direct,
    )
    return ds


@pytest.fixture(scope='module')
def default_mib_raw():
    from rsciio.quantumdetector import load_mib_data
    data = load_mib_data(MIB_TESTDATA_PATH, lazy=True)
    shape = (32, 32, 256, 256)
    return data.reshape(shape)


def test_glob(tmp_path, naughty_filename):
    naughty_dir = tmp_path / naughty_filename
    naughty_dir.mkdir()

    naughty_hdr = naughty_dir / (naughty_filename + '.hdr')
    naughty_hdr.touch()

    naughty_mib_0 = naughty_dir / (naughty_filename + '1.mib')
    naughty_mib_1 = naughty_dir / (naughty_filename + '2.mib')
    naughty_mib_0.touch()
    naughty_mib_1.touch()

    filenames_1 = set(get_filenames(naughty_hdr))
    filenames_2 = set(get_filenames(naughty_mib_0))
    print(filenames_1)
    print(filenames_2)
    target = {str(naughty_mib_0), str(naughty_mib_1)}
    assert filenames_1 == target
    assert filenames_2 == target


@needsdata
def test_detect(lt_ctx):
    params = MIBDataSet.detect_params(MIB_TESTDATA_PATH, lt_ctx.executor)["parameters"]
    assert params == {
        "path": MIB_TESTDATA_PATH,
        "nav_shape": (32, 32),
        "sig_shape": (256, 256)
    }


@needsdata
def test_simple_open(default_mib):
    assert tuple(default_mib.shape) == (32, 32, 256, 256)


@needsdata
def test_positive_sync_offset(default_mib, lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = MIBDataSet(
        path=MIB_TESTDATA_PATH, nav_shape=(32, 32), sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (16,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)
    for t in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (768, 0, 0)
    assert p.slice.shape[0] == 256

    result = lt_ctx.run_udf(dataset=default_mib, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


@needsdata
def test_negative_sync_offset(default_mib, lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = MIBDataSet(
        path=MIB_TESTDATA_PATH, nav_shape=(32, 32), sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (16,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)
    for t in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (768, 0, 0)
    assert p.slice.shape[0] == 256

    result = lt_ctx.run_udf(dataset=default_mib, udf=udf)
    result = result['intensity'].raw_data[:default_mib._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@needsdata
def test_offset_smaller_than_nav_shape(lt_ctx):
    nav_shape = (32, 32)
    sync_offset = -1030

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mib",
            path=MIB_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-1024, 1024\), which is \(-image_count, image_count\)"
        )


@needsdata
def test_offset_greater_than_nav_shape(lt_ctx):
    nav_shape = (32, 32)
    sync_offset = 1030

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mib",
            path=MIB_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-1024, 1024\), which is \(-image_count, image_count\)"
        )


@needsdata
@pytest.mark.with_numba
def test_read(default_mib):
    partitions = default_mib.get_partitions()
    p = next(partitions)
    assert len(p.shape) == 3
    assert tuple(p.shape[1:]) == (256, 256)

    tileshape = Shape(
        (3,) + tuple(default_mib.shape.sig),
        sig_dims=default_mib.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_mib.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    # we get 3D tiles here, because MIB partitions are inherently 3D
    assert tuple(t.tile_slice.shape) == (3, 256, 256)
    for _ in tiles:
        pass


@needsdata
def test_scheme_too_large(default_mib):
    partitions = default_mib.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_mib.shape.sig),
        sig_dims=default_mib.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_mib.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_mib.shape.sig)
    for _ in tiles:
        pass


@needsdata
@pytest.mark.with_numba
def test_read_ahead(default_mib_readahead):
    partitions = default_mib_readahead.get_partitions()
    p = next(partitions)
    assert len(p.shape) == 3
    assert tuple(p.shape[1:]) == (256, 256)

    tileshape = Shape(
        (3,) + tuple(default_mib_readahead.shape.sig),
        sig_dims=default_mib_readahead.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_mib_readahead.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    # we get 3D tiles here, because MIB partitions are inherently 3D
    assert tuple(t.tile_slice.shape) == (3, 256, 256)
    for _ in tiles:
        pass


@needsdata
@pytest.mark.skipif(rsciio is None, reason="No rosettasciio found")
def test_comparison(default_mib, default_mib_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_mib_raw, (-1, *tuple(default_mib.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_mib)


@needsdata
@pytest.mark.skipif(rsciio is None, reason="No rosettasciio found")
def test_comparison_roi(default_mib, default_mib_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_mib.shape.nav),
        p=[0.1, 0.9]
    )
    udf = ValidationUDF(reference=default_mib_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_mib, roi=roi)


@needsdata
def test_pickle_is_small(default_mib):
    pickled = cloudpickle.dumps(default_mib)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 2 * 1024


@needsdata
def test_apply_mask_analysis(default_mib, lt_ctx):
    mask = np.ones((256, 256))
    analysis = lt_ctx.create_mask_analysis(factories=[lambda: mask], dataset=default_mib)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (32, 32)


@needsdata
def test_sum_analysis(default_mib, lt_ctx):
    analysis = lt_ctx.create_sum_analysis(dataset=default_mib)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (256, 256)


@needsdata
def test_pick_analysis(default_mib, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_mib, parameters={"x": 16, "y": 16})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (256, 256)


@needsdata
@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.slow
def test_correction(default_mib, lt_ctx, with_roi):
    ds = default_mib

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx, exclude=[(45, 144), (124, 30)])
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@needsdata
@pytest.mark.with_numba
def test_with_roi(default_mib, lt_ctx):
    udf = PickUDF()
    roi = np.zeros(default_mib.shape.nav, dtype=bool)
    roi[0] = 1
    res = lt_ctx.run_udf(udf=udf, dataset=default_mib, roi=roi)
    np.array(res['intensity']).shape == (1, 256, 256)


@needsdata
def test_read_at_boundaries(default_mib, lt_ctx):
    nav_shape = (32, 32)
    ds_odd = MIBDataSet(path=MIB_TESTDATA_PATH, nav_shape=nav_shape)
    ds_odd = ds_odd.initialize(lt_ctx.executor)

    sumjob_odd = lt_ctx.create_sum_analysis(dataset=ds_odd)
    res_odd = lt_ctx.run(sumjob_odd)

    sumjob = lt_ctx.create_sum_analysis(dataset=default_mib)
    res = lt_ctx.run(sumjob)

    assert np.allclose(res[0].raw_data, res_odd[0].raw_data)


@needsdata
def test_diagnostics(default_mib):
    print(default_mib.diagnostics)


@needsdata
def test_cache_key_json_serializable(default_mib):
    json.dumps(default_mib.get_cache_key())


@needsdata
@pytest.mark.dist
def test_mib_dist(dist_ctx):
    nav_shape = (32, 32)
    ds = MIBDataSet(path=MIB_TESTDATA_PATH, nav_shape=nav_shape)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    results = dist_ctx.run(analysis)
    assert results[0].raw_data.shape == (256, 256)


@needsdata
def test_too_many_files(lt_ctx):
    ds = MIBDataSet(path=MIB_TESTDATA_PATH, nav_shape=(32, 32))

    with mock.patch('libertem.io.dataset.mib.glob', side_effect=lambda p: [
            "/a/%d.mib" % i
            for i in range(256*256)
    ]):
        with pytest.warns(RuntimeWarning) as record:
            ds._filenames()

    assert len(record) == 1
    assert "Saving data in many small files" in record[0].message.args[0]


@needsdata
def test_not_too_many_files(lt_ctx):
    ds = MIBDataSet(path=MIB_TESTDATA_PATH, nav_shape=(32, 32))

    with mock.patch('libertem.io.dataset.mib.glob', side_effect=lambda p: [
            "/a/%d.mib" % i
            for i in range(256)
    ]):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            ds._filenames()


@needsdata
def test_reshape_nav(lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("mib", path=MIB_TESTDATA_PATH, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("mib", path=MIB_TESTDATA_PATH, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("mib", path=MIB_TESTDATA_PATH, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


@needsdata
def test_incorrect_sig_shape(lt_ctx):
    nav_shape = (32, 32)
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "mib",
            path=MIB_TESTDATA_PATH,
            nav_shape=nav_shape,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 65536"
    )


@needsdata
def test_scan_size_deprecation(lt_ctx):
    scan_size = (2, 2)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "mib",
            path=MIB_TESTDATA_PATH,
            scan_size=scan_size,
        )
    assert tuple(ds.shape) == (2, 2, 256, 256)


@needsdata
def test_compare_backends(lt_ctx, default_mib, buffered_mib):
    y = random.choice(range(default_mib.shape.nav[0]))
    x = random.choice(range(default_mib.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_mib,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_mib,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_mib, direct_mib):
    y = random.choice(range(default_mib.shape.nav[0]))
    x = random.choice(range(default_mib.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_mib,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_mib,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_mib, buffered_mib, as_sparse):
    roi = np.zeros(default_mib.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[16] = True
    roi[32] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_mib, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_mib, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.slow
def test_run_many_files(lt_ctx):
    # import tarfile
    # archive = pathlib.Path(MIB_TESTDATA_PATH).parent / 'many_mib.tar.gz'
    # outdir = pathlib.Path(tmpdir_factory.mktemp('many_mib_files'))

    # with tarfile.open(archive, mode='r:gz') as tp:
    #     tp.extractall(path=outdir)

    out_dir = pathlib.Path(MIB_TESTDATA_PATH).parent / 'many_mib'
    hdr_path = out_dir / 'w_140 h_139-2map.hdr'
    nav_shape = (139, 141)
    lt_ctx.load('mib', path=hdr_path, nav_shape=nav_shape)


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("mib", MIB_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        if 'nav_shape' not in params:
            params['nav_shape'] = (1, 1)
        ds_params_tester(*args, **params)



import libertem
import libertem.common.slice
import libertem.executor.dask
import libertem.executor.inline
import libertem.io.dataset.hdf5
import libertem.io.dataset.mib
import libertem.io.dataset.raw
# import libertem.io.ingest.cli
# import libertem.io.ingest.empad
# import libertem.io.ingest.hdf5
# import libertem.io.ingest.sink
import libertem.web.server
import libertem.web.cli
import libertem.api
import libertem.preload
import libertem.udf  # NOQA: F401

# old imports, refs #1031
from libertem.corrections import CorrectionSet  # NOQA: F401
from libertem.corrections.corrset import CorrectionSet  # NOQA: F401,F811
from libertem.corrections.detector import (  # NOQA: F401
    correct, RepairDescriptor, RepairValueError, correct_dot_masks
)

from libertem.masks import to_dense, to_sparse, is_sparse  # NOQA: F401

from libertem.viz import encode_image  # NOQA: F401
from libertem.viz.base import encode_image  # NOQA: F401,F811


def test_stuff():
    print("stuff")
